import attr
from .. import repositories, entities, exceptions


@attr.s
class Task:
    """
    Task object
    """

    # platform
    name = attr.ib()
    status = attr.ib()
    project_id = attr.ib()
    metadata = attr.ib(repr=False)
    id = attr.ib()
    url = attr.ib(repr=False)
    task_owner = attr.ib(repr=False)
    creator = attr.ib()
    due_date = attr.ib()
    dataset_id = attr.ib()
    spec = attr.ib()
    recipe_id = attr.ib(repr=False)
    query = attr.ib(repr=False)
    assignmentIds = attr.ib(repr=False)

    # sdk
    _client_api = attr.ib(repr=False)
    _current_assignments = attr.ib(default=None, repr=False)
    _assignments = attr.ib(default=None, repr=False)
    _project = attr.ib(default=None, repr=False)
    _dataset = attr.ib(default=None, repr=False)
    _tasks = attr.ib(default=None, repr=False)

    @classmethod
    def from_json(cls, _json, client_api, project=None, dataset=None):
        return cls(
            name=_json.get('name', None),
            status=_json.get('status', None),
            project_id=_json.get('projectId', None),
            metadata=_json.get('metadata', dict()),
            url=_json.get('url', None),
            spec=_json.get('spec', None),
            id=_json['id'],
            creator=_json.get('creator', None),
            due_date=_json.get('dueDate', 0),
            dataset_id=_json.get('datasetId', None),
            recipe_id=_json.get('recipeId', None),
            query=_json.get('query', None),
            task_owner=_json.get('taskOwner', None),
            assignmentIds=_json.get('assignmentIds', list()),
            dataset=dataset,
            project=project,
            client_api=client_api
        )

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        """
        _json = attr.asdict(self, filter=attr.filters.exclude(attr.fields(Task)._client_api,
                                                              attr.fields(Task)._project,
                                                              attr.fields(Task).project_id,
                                                              attr.fields(Task).dataset_id,
                                                              attr.fields(Task).recipe_id,
                                                              attr.fields(Task).task_owner,
                                                              attr.fields(Task).due_date,
                                                              attr.fields(Task)._tasks,
                                                              attr.fields(Task)._dataset,
                                                              attr.fields(Task)._current_assignments,
                                                              attr.fields(Task)._assignments))
        _json['projectId'] = self.project_id
        _json['datasetId'] = self.dataset_id
        _json['recipeId'] = self.recipe_id
        _json['task_owner'] = self.task_owner
        _json['dueDate'] = self.due_date
        return _json

    @property
    def current_assignments(self):
        if self._current_assignments is None:
            self._current_assignments = list()
            for assignment in self.assignmentIds:
                self._current_assignments.append(self._assignments.get(assignment_id=assignment))
        return self._current_assignments

    @property
    def assignments(self):
        if self._assignments is None:
            self._assignments = repositories.Assignments(client_api=self._client_api, dataset=self._dataset,
                                                         project=self.project, task=self, project_id=self.project_id)
        assert isinstance(self._assignments, repositories.Assignments)
        return self._assignments

    @property
    def tasks(self):
        if self._tasks is None:
            self._tasks = repositories.Tasks(client_api=self._client_api, project=self.project, dataset=self.dataset)
        assert isinstance(self._tasks, repositories.Tasks)
        return self._tasks

    @property
    def project(self):
        if self._project is None:
            self.get_project()
            if self._project is None:
                raise exceptions.PlatformException(error='2001',
                                                   message='Missing entity "project". need to "get_project()" ')
        assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def dataset(self):
        if self._dataset is None:
            self.get_dataset()
            if self._dataset is None:
                raise exceptions.PlatformException(error='2001',
                                                   message='Missing entity "dataset". need to "get_dataset()" ')
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    def get_project(self):
        if self._project is None:
            self._project = repositories.Projects(client_api=self._client_api).get(project_id=self.project_id)

    def get_dataset(self):
        if self._dataset is None:
            self._dataset = repositories.Datasets(client_api=self._client_api, project=self._project).get(
                dataset_id=self.dataset_id)

    def update(self, system_metadata=False):
        return self.tasks.update(task=self, system_metadata=system_metadata)

    def create_qa_task(self, assignee_ids):
        return self.tasks.create_qa_task(task=self, assignee_ids=assignee_ids)

    def create_assignment(self, assignment_name, assignee_id, items=None, filters=None):
        """

        :param assignment_name:
        :param assignee_id:
        :param items:
        :param filters:
        :return:
        """
        assignment = self.assignments.create(assignment_name=assignment_name,
                                             assignee_id=assignee_id,
                                             project_id=self.project_id,
                                             filters=filters,
                                             items=items,
                                             dataset=self.dataset)

        assignment.metadata['system']['taskId'] = self.id
        assignment.update(system_metadata=True)
        self.assignmentIds.append(assignment.id)
        self.update()
        self.add_items(filters=filters, items=items)
        return assignment

    def add_items(self, filters=None, items=None, workload=None, limit=0):
        """

        :param limit:
        :param workload:
        :param filters:
        :param items:
        :return:
        """
        return self.tasks.add_items(task=self, filters=filters, items=items, workload=workload, limit=limit)

    def remove_items(self, filters=None, items=None):
        """

        :param filters:
        :param items:
        :return:
        """
        return self.tasks.remove_items(dataset=self.dataset, task=self, filters=filters, items=items)

    def get_items(self, filters=None):
        """

        :return:
        """
        return self.tasks.get_items(task_id=self.id, dataset=self.dataset, filters=filters)



from copy import deepcopy
import traceback
import logging
import json
import jwt
import os

from .. import entities, exceptions, miscellaneous, services

logger = logging.getLogger(name=__name__)


class Annotations:
    """
        Annotations repository
    """

    def __init__(self, client_api: services.ApiClient, item=None, dataset=None, dataset_id=None):
        self._client_api = client_api
        self._item = item
        self._dataset = dataset
        if dataset_id is None:
            if dataset is not None:
                dataset_id = dataset.id
            elif item is not None:
                dataset_id = item.datasetId
        self._dataset_id = dataset_id

    ############
    # entities #
    ############
    @property
    def dataset(self):
        if self._dataset is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "dataset". need to set a Dataset entity or use dataset.annotations repository')
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    @dataset.setter
    def dataset(self, dataset: entities.Dataset):
        if not isinstance(dataset, entities.Dataset):
            raise ValueError('Must input a valid Dataset entity')
        self._dataset = dataset

    @property
    def item(self):
        if self._item is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "item". need to set a Item entity or use item.annotations repository')
        assert isinstance(self._item, entities.Item)
        return self._item

    @item.setter
    def item(self, item: entities.Item):
        if not isinstance(item, entities.Item):
            raise ValueError('Must input a valid Item entity')
        self._item = item

    ###########
    # methods #
    ###########
    def get(self, annotation_id):
        """
            Get a single annotation

        :param annotation_id:
        :return: Annotation object or None
        """
        success, response = self._client_api.gen_request(req_type='get',
                                                         path='/annotations/{}'.format(annotation_id))
        if success:
            annotation = entities.Annotation.from_json(_json=response.json(),
                                                       annotations=self,
                                                       item=self._item)
        else:
            raise exceptions.PlatformException(response)
        return annotation

    def _build_entities_from_response(self, response_items):
        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(response_items))]
        # return triggers list
        for i_json, _json in enumerate(response_items):
            jobs[i_json] = pool.apply_async(entities.Annotation._protected_from_json,
                                            kwds={'client_api': self._client_api,
                                                  '_json': _json,
                                                  'item': self._item,
                                                  'annotations': self})
        # wait for all jobs
        _ = [j.wait() for j in jobs]
        # get all results
        results = [j.get() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        return miscellaneous.List([r[1] for r in results if r[0] is True])

    def _list(self, filters: entities.Filters):
        """
        Get dataset items list This is a browsing endpoint, for any given path item count will be returned,
        user is expected to perform another request then for every folder item to actually get the its item list.

        :param filters: Filters entity or a dictionary containing filters parameters
        :return: json response
        """
        # prepare request
        success, response = self._client_api.gen_request(req_type="POST",
                                                         path="/datasets/{}/query".format(self._dataset_id),
                                                         json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    def list(self, filters: entities.Filters = None, page_offset=None, page_size=None):
        """
        List Annotation

        :param filters: Filters entity or a dictionary containing filters parameters
        :param page_offset:
        :param page_size:
        :return: Pages object
        """
        if self._dataset_id is not None:
            if filters is None:
                filters = entities.Filters(resource=entities.FiltersResource.ANNOTATION)

            if self._item is not None and not filters.has_field('itemId'):
                filters = deepcopy(filters)
                filters.page_size = 1000
                filters.add(field='itemId', values=self.item.id, method=entities.FiltersMethod.AND)

            # assert type filters
            if not isinstance(filters, entities.Filters):
                raise exceptions.PlatformException('400', 'Unknown filters type')

            # page size
            if page_size is None:
                # take from default
                page_size = filters.page_size
            else:
                filters.page_size = page_size

            # page offset
            if page_offset is None:
                # take from default
                page_offset = filters.page
            else:
                filters.page = page_offset

            paged = entities.PagedEntities(items_repository=self,
                                           filters=filters,
                                           page_offset=page_offset,
                                           page_size=page_size,
                                           client_api=self._client_api)
            paged.get_page()

            if self._item is not None:
                if paged.total_pages_count > 1:
                    annotations = list()
                    for page in paged:
                        annotations += page
                else:
                    annotations = paged.items
                return entities.AnnotationCollection(annotations=annotations, item=self._item)
            else:
                return paged
        else:
            raise exceptions.PlatformException('400',
                                               'Please use item.annotations.list() or dataset.annotations.list() '
                                               'to perform this action.')

    def show(self, image=None, thickness=1, with_text=False, height=None, width=None,
             annotation_format=entities.ViewAnnotationOptions.MASK):
        """
        Show annotations

        :param image: empty or image to draw on
        :param height: height
        :param width: width
        :param thickness: line thickness
        :param with_text: add label to annotation
        :param annotation_format: options: dl.ViewAnnotationOptions.list()
        :return: ndarray of the annotations
        """
        # get item's annotations
        annotations = self.list()

        return annotations.show(image=image,
                                width=width,
                                height=height,
                                thickness=thickness,
                                with_text=with_text,
                                annotation_format=annotation_format)

    def download(self, filepath, annotation_format=entities.ViewAnnotationOptions.MASK, height=None, width=None,
                 thickness=1, with_text=False):
        """
            Save annotation format to file

        :param filepath: Target download directory
        :param annotation_format: optional - dl.ViewAnnotationOptions.list()
        :param height: optional - image height
        :param width: optional - image width
        :param thickness: optional - annotation format, default =1
        :param with_text: optional - draw annotation with text, default = False
        :return:
        """
        # get item's annotations
        annotations = self.list()

        # height/weight
        if height is None:
            if self.item.height is None:
                raise exceptions.PlatformException('400', 'Height must be provided')
            height = self.item.height
        if width is None:
            if self.item.width is None:
                raise exceptions.PlatformException('400', 'Width must be provided')
            width = self.item.width

        return annotations.download(filepath=filepath,
                                    width=width,
                                    height=height,
                                    thickness=thickness,
                                    with_text=with_text,
                                    annotation_format=annotation_format)

    def _delete_single_annotation(self, w_annotation_id):
        try:

            creator = jwt.decode(self._client_api.token, algorithms=['HS256'], verify=False)['email']
            payload = {'username': creator}
            success, response = self._client_api.gen_request(req_type='delete',
                                                             path='/annotations/{}'.format(w_annotation_id),
                                                             json_req=payload)

            if not success:
                raise exceptions.PlatformException(response)
            status = True
        except Exception:
            status = False
            response = traceback.format_exc()
        return status, response

    def delete(self, annotation=None, annotation_id=None):
        """
            Remove an annotation from item

        :param annotation: Annotation object
        :param annotation_id: annotation id
        :return: True/False
        """
        if annotation_id is not None:
            pass
        elif annotation is not None and isinstance(annotation, entities.Annotation):
            annotation_id = annotation.id
        elif annotation is not None and isinstance(annotation, str) and annotation.lower() == 'all':
            annotation_id = [annotation.id for annotation in self.list()]
        else:
            raise exceptions.PlatformException(error='400', message='Must input annotation id or annotation entity')
        # get creator from token

        if not isinstance(annotation_id, list):
            annotation_id = [annotation_id]

        pool = self._client_api.thread_pools(pool_name='annotation.update')
        jobs = [None for _ in range(len(annotation_id))]
        for i_ann, ann_id in enumerate(annotation_id):
            jobs[i_ann] = pool.apply_async(func=self._delete_single_annotation,
                                           kwds={'w_annotation_id': ann_id})
        # wait for jobs to be finish
        _ = [j.wait() for j in jobs]
        # get all results
        results = [j.get() for j in jobs]
        out_annotations = [r[1] for r in results if r[0] is True]
        out_errors = [r[1] for r in results if r[0] is False]
        if len(out_errors) == 0:
            logger.debug('Annotation/s delete successfully. {}/{}'.format(len(out_annotations), len(results)))
        else:
            logger.error(out_errors)
            logger.error('Annotation/s delete with {} errors'.format(len(out_errors)))
        return True

    def _update_single_annotation(self, w_annotation, system_metadata):
        try:
            if isinstance(w_annotation, entities.Annotation):
                annotation_id = w_annotation.id
                annotation = w_annotation.to_json()
            else:
                raise exceptions.PlatformException('400',
                                                   'unknown annotations type: {}'.format(type(w_annotation)))

            url_path = '/annotations/{}'.format(annotation_id)
            if system_metadata:
                url_path += '?system=true'
            suc, response = self._client_api.gen_request(req_type='put',
                                                         path=url_path,
                                                         json_req=annotation)
            if suc:
                result = entities.Annotation.from_json(_json=response.json(),
                                                       annotations=self,
                                                       item=self._item)
            else:
                raise exceptions.PlatformException(response)
            status = True
        except Exception:
            status = False
            result = traceback.format_exc()
        return status, result

    def update(self, annotations, system_metadata=False):
        """
            Update an existing annotation.

        :param annotations:
        :param system_metadata:
        :return: True
        """
        pool = self._client_api.thread_pools(pool_name='annotation.update')
        if not isinstance(annotations, list):
            annotations = [annotations]
        jobs = [None for _ in range(len(annotations))]
        for i_ann, ann in enumerate(annotations):
            jobs[i_ann] = pool.apply_async(func=self._update_single_annotation,
                                           kwds={'w_annotation': ann,
                                                 'system_metadata': system_metadata})
        # wait for jobs to be finish
        _ = [j.wait() for j in jobs]
        # get all results
        results = [j.get() for j in jobs]
        out_annotations = [r[1] for r in results if r[0] is True]
        out_errors = [r[1] for r in results if r[0] is False]
        if len(out_errors) == 0:
            logger.debug('Annotation/s updated successfully. {}/{}'.format(len(out_annotations), len(results)))
        else:
            logger.error(out_errors)
            logger.error('Annotation/s updated with {} errors'.format(len(out_errors)))
        return out_annotations

    def _upload_annotations(self, annotations):
        annotations_list = list()
        for annotation in annotations:
            if isinstance(annotation, str):
                annotation = json.loads(annotation)
            elif isinstance(annotation, entities.Annotation):
                annotation = annotation.to_json()
            elif isinstance(annotation, dict):
                annotation = annotation
            else:
                raise exceptions.PlatformException('400',
                                                   'unknown annotations type: {}'.format(type(annotation)))
            annotations_list.append(annotation)
        suc, response = self._client_api.gen_request(req_type='post',
                                                     path='/items/{}/annotations'.format(self.item.id),
                                                     json_req=annotations_list)
        if suc:
            return_annotations = response.json()
            if not isinstance(return_annotations, list):
                return_annotations = [return_annotations]
            result = entities.AnnotationCollection.from_json(_json=return_annotations,
                                                             item=self.item)
        else:
            raise exceptions.PlatformException(response)
        return result

    def upload(self, annotations):
        """
        Create a new annotation

        :param annotations: list or single annotation of type Annotation
        :return: list of annotation objects
        """
        # make list if not list
        if isinstance(annotations, entities.AnnotationCollection):
            annotations = annotations.annotations
        if not isinstance(annotations, list):
            annotations = [annotations]
            if isinstance(annotations[0], str) and os.path.isfile(annotations[0]):
                with open(annotations[0], 'r', encoding="utf8") as f:
                    annotations = json.load(f)
                if isinstance(annotations, dict):
                    if 'annotations' in annotations:
                        annotations = annotations['annotations']
                    elif 'data' in annotations:
                        annotations = annotations['data']
                    else:
                        exceptions.PlatformException('400', 'Unknown annotation file format')
        out_annotations = self._upload_annotations(annotations=annotations)
        return out_annotations

    def update_status(self, annotation: entities.Annotation = None, annotation_id=None, status='issue'):
        """
        Set status on annotation
        :param annotation: optional - Annotation entity
        :param annotation_id: optional - annotation id to set status
        :param status: can be "issue", "approved", "review"
        """
        if annotation is None:
            if annotation_id is None:
                raise ValueError('must input on of "annotation" or "annotation_id"')
            annotation = self.get(annotation_id=annotation_id)
        if status not in ["issue", "approved", "review"]:
            raise ValueError('status must be on of: "issue", "approved", "review"')
        annotation.status = status
        annotation.update(system_metadata=True)

    def builder(self):
        return entities.AnnotationCollection(item=self.item)

    ##################
    # async function #
    ##################
    async def _async_upload_annotations(self, annotations):
        async with self._client_api.event_loops('items.upload').semaphore('annotations.upload'):
            annotations_list = list()
            for annotation in annotations:
                if isinstance(annotation, str):
                    annotation = json.loads(annotation)
                elif isinstance(annotation, entities.Annotation):
                    annotation = annotation.to_json()
                elif isinstance(annotation, dict):
                    annotation = annotation
                else:
                    raise exceptions.PlatformException(error='400',
                                                       message='unknown annotations type: {}'.format(
                                                           type(annotation)))
                annotations_list.append(annotation)
            success, response = await self._client_api.gen_async_request(
                req_type='post',
                path='/items/{}/annotations'.format(self.item.id),
                json_req=annotations_list)
            if success:
                return_annotations = response.json()
                if not isinstance(return_annotations, list):
                    return_annotations = [return_annotations]
                result = entities.AnnotationCollection.from_json(_json=return_annotations,
                                                                 item=self.item)
            else:
                raise exceptions.PlatformException(response)
            return result



import copy
import mimetypes
from itertools import groupby
from multiprocessing.pool import ThreadPool
import os
import json
import tqdm
from PIL import Image
from .. import exceptions, entities, utilities
from jinja2 import Environment, PackageLoader
import numpy as np
import logging
import xml.etree.ElementTree as Et
from ..services import Reporter
import traceback

logger = logging.getLogger(name=__name__)


class AnnotationFormat:
    YOLO = 'yolo'
    COCO = 'coco'
    VOC = 'voc'
    DATALOOP = 'dataloop'


class COCOUtils:

    @staticmethod
    def binary_mask_to_rle(binary_mask, height, width):
        rle = {'counts': [], 'size': [height, width]}
        counts = rle.get('counts')
        for i, (value, elements) in enumerate(groupby(binary_mask.ravel(order='F'))):
            if i == 0 and value == 1:
                counts.append(0)
            counts.append(len(list(elements)))
        return rle

    @staticmethod
    def polygon_to_rle(geo, height, width):
        segmentation = [float(n) for n in geo.flatten()]
        area = np.sum(entities.Segmentation.from_polygon(geo=geo, label=None, shape=(height, width)).geo > 0)
        return [segmentation], int(area)

    @staticmethod
    def rle_to_binary_mask(rle):
        rows, cols = rle['size']
        rle_numbers = rle['counts']

        if len(rle_numbers) % 2 != 0:
            rle_numbers.append(0)

        rle_pairs = np.array(rle_numbers).reshape(-1, 2)
        img = np.zeros(rows * cols, dtype=np.uint8)
        index = 0
        for i, length in rle_pairs:
            index += i
            img[index:index + length] = 1
            index += length
        img = img.reshape(cols, rows)
        return img.T

    @staticmethod
    def rle_to_binary_polygon(segmentation):
        return [segmentation[x:x + 2] for x in range(0, len(segmentation), 2)]


class Converter:
    """
    Annotation Converter
    """

    def __init__(self):
        self.known_formats = [AnnotationFormat.YOLO,
                              AnnotationFormat.COCO,
                              AnnotationFormat.VOC,
                              AnnotationFormat.DATALOOP]
        self.converter_dict = {
            AnnotationFormat.YOLO: {"from": self.from_yolo, "to": self.to_yolo},
            AnnotationFormat.COCO: {"from": self.from_coco, "to": self.to_coco},
            AnnotationFormat.VOC: {"from": self.from_voc, "to": self.to_voc},
        }
        self.dataset = None
        self.save_to_format = None
        self.xml_template_path = 'voc_annotation_template.xml'
        self.labels = dict()

    def _get_labels(self):
        if self.dataset:
            self.labels = {label.tag: i_label for i_label, label in enumerate(self.dataset.labels)}

    def convert_dataset(self,
                        dataset,
                        to_format,
                        local_path,
                        conversion_func=None,
                        filters=None,
                        annotation_filter=None):
        """
        Convert entire dataset

        :param annotation_filter:
        :param dataset:
        :param to_format:
        :param local_path:
        :param conversion_func: Custom conversion service
        :param filters: optional
        :return:
        """
        if to_format.lower() == AnnotationFormat.COCO:
            return self.__convert_dataset_to_coco(dataset=dataset,
                                                  local_path=local_path,
                                                  filters=filters,
                                                  annotation_filter=annotation_filter)
        num_workers = 6
        assert isinstance(dataset, entities.Dataset)
        self.dataset = dataset

        # download annotations
        if annotation_filter is None:
            dataset.download_annotations(local_path=local_path, overwrite=True)
        local_annotations_path = os.path.join(local_path, "json")
        output_annotations_path = os.path.join(local_path, to_format)
        pool = ThreadPool(processes=num_workers)
        i_item = 0
        pages = dataset.items.list(filters=filters)

        # if yolo - create labels file
        if to_format == AnnotationFormat.YOLO:
            self._get_labels()
            labels = [label.tag for label in dataset.labels]
            with open('{}/{}.names'.format(local_path, dataset.name), 'w') as fp:
                for label in labels:
                    fp.write("{}\n".format(label))

        pbar = tqdm.tqdm(total=pages.items_count)
        reporter = Reporter(num_workers=pages.items_count, resource=Reporter.CONVERTER)
        for page in pages:
            for item in page:
                # create input annotations json
                in_filepath = os.path.join(local_annotations_path, item.filename[1:])
                name, ext = os.path.splitext(in_filepath)
                in_filepath = name + '.json'

                save_to = os.path.dirname(in_filepath.replace(local_annotations_path, output_annotations_path))

                if not os.path.isdir(save_to):
                    os.makedirs(save_to, exist_ok=True)

                converter = utilities.Converter()
                converter.dataset = self.dataset
                converter.labels = self.labels
                converter.save_to_format = self.save_to_format
                converter.xml_template_path = self.xml_template_path

                pool.apply_async(
                    func=self.__save_filtered_annotations_and_convert,
                    kwds={
                        "to_format": to_format,
                        "from_format": AnnotationFormat.DATALOOP,
                        "file_path": in_filepath,
                        "save_locally": True,
                        "save_to": save_to,
                        'conversion_func': conversion_func,
                        'item': item,
                        'pbar': pbar,
                        'filters': annotation_filter,
                        'reporter': reporter,
                        'i_item': i_item
                    }
                )
                i_item += 1
        pool.close()
        pool.join()
        pool.terminate()
        pbar.close()

        if reporter.has_errors:
            log_filepath = reporter.generate_log_files()
            logger.warning(
                'Converted with some errors. Please see log in {} for more information.'.format(log_filepath))

        logger.info('Total converted: {}'.format(reporter.status_count('success')))
        logger.info('Total skipped: {}'.format(reporter.status_count('skip')))
        logger.info('Total failed: {}'.format(reporter.status_count('failed')))

    def __save_filtered_annotations_and_convert(self, item, filters, to_format, from_format, file_path,
                                                save_locally=False,
                                                save_to=None, conversion_func=None,
                                                pbar=None, **kwargs):
        reporter = kwargs.get('reporter', None)
        i_item = kwargs.get('i_item', None)

        try:
            if item.annotated:
                if filters is not None:
                    assert filters.resource == entities.FiltersResource.ANNOTATION
                    copy_filters = copy.deepcopy(filters)
                    copy_filters.add(field='itemId', values=item.id, method='and')
                    annotations_page = item.dataset.items.list(filters=copy_filters)
                    annotations = item.annotations.builder()
                    for page in annotations_page:
                        for annotation in page:
                            annotations.annotations.append(annotation)

                    if not os.path.isdir(os.path.dirname(file_path)):
                        os.makedirs(os.path.dirname(file_path))
                    with open(file_path, 'w') as f:
                        json.dump(annotations.to_json(), f, indent=2)

                converted_annotations = self.convert_file(item=item,
                                                          to_format=to_format,
                                                          from_format=from_format,
                                                          file_path=file_path,
                                                          save_locally=save_locally,
                                                          save_to=save_to,
                                                          conversion_func=conversion_func,
                                                          pbar=pbar)

                success, errors = self._sort_annotations(annotations=converted_annotations)

                if not success:
                    raise Exception('Failed to convert item annotations: \n{}'.format(errors))
                elif errors:
                    raise Exception('Partial conversion: \n{}'.format(errors))

                if reporter is not None and i_item is not None:
                    reporter.set_index(i_item=i_item, status='success', success=True)
            else:
                if reporter is not None and i_item is not None:
                    reporter.set_index(i_item=i_item, status='skip', success=True)
        except Exception:
            if reporter is not None and i_item is not None:
                reporter.set_index(i_item=i_item, status='failed', success=False, error=traceback.format_exc(),
                                   ref=item.id)

    @staticmethod
    def __gen_coco_categories(instance_map):
        categories = list()
        for label, label_id in instance_map.items():
            label_name, sup = label.split('.')[-1], '.'.join(label.split('.')[0:-1])
            category = {'id': label_id, 'name': label_name}
            if sup:
                category['supercategory'] = sup
            categories.append(category)
        return categories

    def __convert_dataset_to_coco(self, dataset: entities.Dataset, local_path, filters=None, annotation_filter=None):
        pages = dataset.items.list(filters=filters)
        dataset.download_annotations(local_path=local_path)
        path_to_dataloop_annotations_dir = os.path.join(local_path, 'json')
        label_to_id = dataset.instance_map
        categories = self.__gen_coco_categories(instance_map=label_to_id)
        images = [None for _ in range(pages.items_count)]
        converted_annotations = [None for _ in range(pages.items_count)]
        item_id_counter = 0
        pool = ThreadPool(processes=11)
        pbar = tqdm.tqdm(total=pages.items_count)
        reporter = Reporter(num_workers=pages.items_count, resource=Reporter.CONVERTER)
        for page in pages:
            for item in page:
                pool.apply_async(func=self.__single_item_to_coco,
                                 kwds={
                                     'item': item,
                                     'images': images,
                                     'path_to_dataloop_annotations_dir': path_to_dataloop_annotations_dir,
                                     'item_id': item_id_counter,
                                     'reporter': reporter,
                                     'converted_annotations': converted_annotations,
                                     'annotation_filter': annotation_filter,
                                     'label_to_id': label_to_id,
                                     'pbar': pbar
                                 })
                item_id_counter += 1

        pool.close()
        pool.join()
        pool.terminate()

        total_converted_annotations = list()
        for ls in converted_annotations:
            if ls is not None:
                total_converted_annotations += ls

        for i_ann, ann in enumerate(total_converted_annotations):
            ann['id'] = i_ann

        info = {
            'description': dataset.name
        }

        coco_json = {'images': [image for image in images if image is not None],
                     'info': info,
                     'annotations': total_converted_annotations,
                     'categories': categories}

        with open(os.path.join(local_path, 'coco.json'), 'w+') as f:
            json.dump(coco_json, f)

        if reporter.has_errors:
            log_filepath = reporter.generate_log_files()
            logger.warning(
                'Converted with some errors. Please see log in {} for more information.'.format(log_filepath))

        logger.info('Total converted: {}'.format(reporter.status_count('success')))
        logger.info('Total failed: {}'.format(reporter.status_count('failed')))

        return coco_json

    def __single_item_to_coco(self, item, images, path_to_dataloop_annotations_dir, item_id, converted_annotations,
                              annotation_filter, label_to_id, reporter, pbar=None):
        try:
            images[item_id] = {'file_name': item.name,
                               'id': item_id,
                               'width': item.width,
                               'height': item.height
                               }
            if annotation_filter is None:
                try:
                    filename, ext = os.path.splitext(item.filename)
                    filename = '{}.json'.format(filename[1:])
                    with open(os.path.join(path_to_dataloop_annotations_dir, filename), 'r', encoding="utf8") as f:
                        annotations = json.load(f)['annotations']
                    annotations = entities.AnnotationCollection.from_json(annotations)
                except Exception:
                    annotations = item.annotations.list()
            else:
                copy_filters = copy.deepcopy(annotation_filter)
                copy_filters.add(field='itemId', values=item.id, method='and')
                annotations_page = item.dataset.items.list(filters=copy_filters)
                annotations = item.annotations.builder()
                for page in annotations_page:
                    for annotation in page:
                        annotations.annotations.append(annotation)

            item_converted_annotations = list()
            for i_annotation, annotation in enumerate(annotations.annotations):
                try:
                    ann = self.to_coco(annotation=annotation, item=item)
                    ann['category_id'] = label_to_id[annotation.label]
                    ann['image_id'] = item_id
                    ann['id'] = int('{}{}'.format(item_id, i_annotation))

                    item_converted_annotations.append(ann)
                except Exception:
                    err = 'Error converting annotation: \n' \
                          'Item: {}, annotation: {} - ' \
                          'fail to convert some of the annotation\n{}'.format(item_id,
                                                                              annotation.id,
                                                                              traceback.format_exc())
                    item_converted_annotations.append(err)
            success, errors = self._sort_annotations(annotations=item_converted_annotations)
            converted_annotations[item_id] = success
            if errors:
                reporter.set_index(i_item=item_id, ref=item.id, status='failed', success=False,
                                   error=errors)
            else:
                reporter.set_index(i_item=item_id, status='success', success=True)
        except Exception:
            reporter.set_index(i_item=item_id, ref=item.id, status='failed', success=False,
                               error=traceback.format_exc())
            raise

        if pbar is not None:
            pbar.update()

    def _upload_coco_labels(self, coco_json):
        labels = coco_json.get('categories', None)
        upload_labels = dict()
        for label in labels:
            if 'supercategory' in label and label['supercategory'] is not None:
                if label['supercategory'] not in upload_labels:
                    upload_labels[label['supercategory']] = entities.Label(tag=label['supercategory'])
                upload_labels[label['supercategory']].children.append(entities.Label(tag=label['name']))
                tag = '{}.{}'.format(label['supercategory'], label['name'])
            else:
                tag = label['name']
                upload_labels[label['name']] = entities.Label(tag=tag)
            self.labels[tag] = label['id']

        self.dataset.add_labels(upload_labels.values())

    def _upload_coco_dataset(self, local_items_path, local_annotations_path, only_bbox=False):
        logger.info('loading annotations json...')
        with open(local_annotations_path, 'r', encoding="utf8") as f:
            coco_json = json.load(f)

        try:
            logger.info('Uploading labels to dataset')
            self._upload_coco_labels(coco_json=coco_json)
        except Exception:
            logger.warning('Failed to upload labels to dataset, please add manually')
            self._get_labels()

        image_annotations = dict()
        image_name_id = dict()
        for image in coco_json['images']:
            image_metadata = image
            image_annotations[image['file_name']] = {
                'id': image['id'],
                'metadata': image_metadata,
                'annotations': list()
            }
            image_name_id[image['id']] = image['file_name']
        for ann in coco_json['annotations']:
            image_annotations[image_name_id[ann['image_id']]]['annotations'].append(ann)

        return self._upload_directory(local_items_path=local_items_path,
                                      local_annotations_path=image_annotations,
                                      from_format=AnnotationFormat.COCO,
                                      only_bbox=only_bbox)

    def _read_labels(self, labels_file_path):
        if labels_file_path:
            with open(labels_file_path, 'r') as fp:
                labels = [line.strip() for line in fp.readlines()]
                self.dataset.add_labels(label_list=labels)
                self.labels = {label: i_label for i_label, label in enumerate(labels)}
        else:
            logger.warning('No labels file path provided (.names), skipping labels upload')
            self._get_labels()

    def _upload_yolo_dataset(self, local_items_path, local_annotations_path, labels_file_path):
        self._read_labels(labels_file_path=labels_file_path)

        return self._upload_directory(local_items_path=local_items_path,
                                      local_annotations_path=local_annotations_path,
                                      from_format=AnnotationFormat.YOLO)

    def _upload_voc_dataset(self, local_items_path, local_annotations_path, **_):
        # TODO - implement VOC annotations upload
        logger.warning('labels upload from VOC dataset is not implemented, please upload labels manually')
        return self._upload_directory(local_items_path=local_items_path,
                                      local_annotations_path=local_annotations_path,
                                      from_format=AnnotationFormat.VOC)

    def _upload_directory(self, local_items_path, local_annotations_path, from_format, conversion_func=None, **kwargs):
        """
        Convert annotation files in entire directory

        :param local_items_path:
        :param local_annotations_path:
        :param from_format:
        :param conversion_func:
        :return:
        """
        only_bbox = kwargs.get('only_bbox', False)
        file_count = sum(len([file for file in files if not file.endswith('.xml')]) for _, _, files in
                         os.walk(local_items_path))

        reporter = Reporter(num_workers=file_count, resource=Reporter.CONVERTER)
        pbar = tqdm.tqdm(total=file_count)

        pool = ThreadPool(processes=6)
        i_item = 0
        metadata = None
        for path, subdirs, files in os.walk(local_items_path):
            for name in files:
                if not os.path.isfile(os.path.join(path, name)):
                    continue
                if from_format == AnnotationFormat.COCO:
                    item_filepath = os.path.join(path, name)
                    ann_filepath = local_annotations_path[name]
                    metadata = {'user': local_annotations_path[name]['metadata']}
                elif from_format == AnnotationFormat.VOC:
                    if name.endswith('.xml'):
                        continue
                    else:
                        ext = os.path.splitext(name)[-1]
                        try:
                            m = mimetypes.types_map[ext.lower()]
                        except Exception:
                            m = ''

                        if ext == '' or ext is None or 'image' not in m:
                            continue

                    item_filepath = os.path.join(path, name)
                    ann_filepath = os.path.join(path, '.'.join(name.split('.')[0:-1] + ['xml'])).replace(
                        local_items_path, local_annotations_path)
                elif from_format == AnnotationFormat.YOLO:
                    item_filepath = os.path.join(path, name)
                    ann_filepath = os.path.join(path, os.path.splitext(name)[0]) + '.txt'
                    ann_filepath = ann_filepath.replace(local_items_path, local_annotations_path)
                converter = utilities.Converter()
                converter.dataset = self.dataset
                converter.labels = self.labels
                if only_bbox:
                    converter._only_bbox = True
                pool.apply_async(
                    func=converter._upload_item_and_convert,
                    kwds={
                        "from_format": from_format,
                        "item_path": item_filepath,
                        "ann_path": ann_filepath,
                        'conversion_func': conversion_func,
                        "reporter": reporter,
                        'i_item': i_item,
                        'pbar': pbar,
                        'metadata': metadata
                    }
                )
                i_item += 1
        pool.close()
        pool.join()
        pool.terminate()

        if reporter.has_errors:
            log_filepath = reporter.generate_log_files()
            logger.warning(
                'Converted with some errors. Please see log in {} for more information.'.format(log_filepath))

        logger.info('Total converted and uploaded: {}'.format(reporter.status_count('success')))
        logger.info('Total failed: {}'.format(reporter.status_count('failed')))

    def _upload_item_and_convert(self, item_path, ann_path, from_format, conversion_func=None, **kwargs):
        reporter = kwargs.get('reporter', None)
        i_item = kwargs.get('i_item', None)
        pbar = kwargs.get('pbar', None)
        metadata = kwargs.get('metadata', None)

        try:
            item = self.dataset.items.upload(local_path=item_path, item_metadata=metadata)
            if from_format == AnnotationFormat.YOLO:
                image = Image.open(item_path)
                width, height = image.size
                item.width = width
                item.height = height
            converted_annotations = self.convert_file(to_format=AnnotationFormat.DATALOOP,
                                                      from_format=from_format,
                                                      item=item,
                                                      file_path=ann_path,
                                                      save_locally=False,
                                                      conversion_func=conversion_func,
                                                      upload=True,
                                                      pbar=pbar)
            success, errors = self._sort_annotations(annotations=converted_annotations)
            if not success:
                raise Exception("Failed to convert item's annotations: {}\n{}".format(item_path, errors))
            if errors:
                if reporter is not None and i_item is not None:
                    reporter.set_index(i_item=i_item, ref=item_path, status='warning', success=False,
                                       error='partial annotations upload: \n{}'.format(errors))
            else:
                if reporter is not None and i_item is not None:
                    reporter.set_index(i_item=i_item, status='success', success=True, ref=item_path)
        except Exception:
            if reporter is not None and i_item is not None:
                reporter.set_index(i_item=i_item, status='failed', success=False, error=traceback.format_exc(),
                                   ref=item_path)

    def upload_local_dataset(self,
                             from_format,
                             dataset,
                             local_items_path=None,
                             local_labels_path=None,
                             local_annotations_path=None,
                             only_bbox=False):
        """
        Convert and Upload local dataset to dataloop platform

        :param only_bbox: only for coco datasets
        :param from_format:
        :param dataset:
        :param local_items_path:
        :param local_annotations_path:
        :param local_labels_path:
        :param local_items_path:
        :return:
        """
        self.dataset = dataset
        if from_format.lower() == AnnotationFormat.COCO:
            self._upload_coco_dataset(local_items_path=local_items_path, local_annotations_path=local_annotations_path,
                                      only_bbox=only_bbox)
        if from_format.lower() == AnnotationFormat.YOLO:
            self._upload_yolo_dataset(local_items_path=local_items_path, local_annotations_path=local_annotations_path,
                                      labels_file_path=local_labels_path)
        if from_format.lower() == AnnotationFormat.VOC:
            self._upload_voc_dataset(local_items_path=local_items_path, local_annotations_path=local_annotations_path,
                                     labels_file_path=local_labels_path)

    def convert_directory(self, local_path, to_format, from_format, dataset, conversion_func=None):
        """
        Convert annotation files in entire directory

        :param local_path:
        :param to_format:
        :param from_format:
        :param conversion_func:
        :param dataset:
        :return:
        """
        file_count = sum(len(files) for _, _, files in os.walk(local_path))
        reporter = Reporter(num_workers=file_count, resource=Reporter.CONVERTER)

        pool = ThreadPool(processes=6)
        i_item = 0
        for path, subdirs, files in os.walk(local_path):
            for name in files:
                save_to = os.path.join(os.path.split(local_path)[0], to_format)
                save_to = path.replace(local_path, save_to)
                if not os.path.isdir(save_to):
                    os.mkdir(save_to)
                file_path = os.path.join(path, name)
                converter = utilities.Converter()
                if dataset is None:
                    converter.dataset = self.dataset
                else:
                    converter.dataset = dataset
                converter.save_to_format = self.save_to_format
                pool.apply_async(
                    func=converter._convert_and_report,
                    kwds={
                        "to_format": to_format,
                        "from_format": from_format,
                        "file_path": file_path,
                        "save_locally": True,
                        "save_to": save_to,
                        'conversion_func': conversion_func,
                        'reporter': reporter,
                        'i_item': i_item
                    }
                )
                i_item += 1
        pool.close()
        pool.join()
        pool.terminate()

        if reporter.has_errors:
            log_filepath = reporter.generate_log_files()
            logger.warning(
                'Converted with some errors. Please see log in {} for more information.'.format(log_filepath))

        logger.info('Total converted and uploaded: {}'.format(reporter.status_count('success')))
        logger.info('Total failed: {}'.format(reporter.status_count('failed')))

    def _convert_and_report(self, to_format, from_format, file_path, save_locally, save_to, conversion_func, reporter,
                            i_item):
        try:
            converted_annotations = self.convert_file(to_format=to_format, from_format=from_format, file_path=file_path,
                                                      save_locally=save_locally, save_to=save_to,
                                                      conversion_func=conversion_func)
            success, errors = self._sort_annotations(annotations=converted_annotations)
            if errors:
                reporter.set_index(i_item=i_item, ref=file_path, status='warning', success=False,
                                   error='partial annotations upload')
            else:
                reporter.set_index(i_item=i_item, status='success', success=True)
        except Exception:
            reporter.set_index(i_item=i_item, ref=file_path, status='failed', success=False,
                               error=traceback.format_exc())
            raise

    @staticmethod
    def _extract_annotations_from_file(file_path, from_format, to_format, item):
        item_id = None
        annotations = None
        if isinstance(file_path, dict):
            annotations = file_path['annotations']
        elif isinstance(file_path, str) and os.path.isfile(file_path):
            with open(file_path, "r") as f:
                if file_path.endswith(".json"):
                    annotations = json.load(f)
                    if from_format.lower() == AnnotationFormat.DATALOOP:
                        if item is None:
                            item_id = annotations.get('_id', annotations.get('id', None))
                        annotations = annotations["annotations"]
                elif from_format.lower() == AnnotationFormat.YOLO:
                    annotations = [[float(param.replace('\n', '')) for param in ann.split(' ')] for ann in
                                   f.readlines()]
                elif file_path.endswith(".xml"):
                    annotations = Et.parse(f)
                    annotations = [e for e in annotations.iter('object')]
                else:
                    raise NotImplementedError('Unknown file format: {}'.format(file_path))
        else:
            raise NotImplementedError('Unknown file_path: {}'.format(file_path))

        return item_id, annotations

    def convert_file(self, to_format, from_format, file_path, save_locally=False, save_to=None, conversion_func=None,
                     item=None, pbar=None, upload=False, **_):
        """
        Convert file containing annotations

        :param to_format:
        :param from_format:
        :param file_path:
        :param pbar:
        :param upload:
        :param save_locally:
        :param save_to:
        :param conversion_func:
        :param item:
        :return:
        """
        item_id, annotations = self._extract_annotations_from_file(to_format=to_format, from_format=from_format,
                                                                   file_path=file_path, item=item)

        converted_annotations = self.convert(
            to_format=to_format,
            from_format=from_format,
            annotations=annotations,
            conversion_func=conversion_func,
            item=item
        )

        success, errors = self._sort_annotations(annotations=converted_annotations)

        if success:
            if save_locally:
                if item_id is not None:
                    item = self.dataset.items.get(item_id=item_id)
                filename = os.path.split(file_path)[-1]
                filename_no_ext = os.path.splitext(filename)[0]
                save_to = os.path.join(save_to, filename_no_ext)
                self.save_to_file(save_to=save_to, to_format=to_format, annotations=success, item=item)
            elif upload and to_format == AnnotationFormat.DATALOOP:
                item.annotations.upload(annotations=success)

        if pbar is not None:
            pbar.update()

        return converted_annotations

    @staticmethod
    def _sort_annotations(annotations):
        errors = list()
        success = list()

        for ann in annotations:
            if isinstance(ann, str) and 'Error converting annotation' in ann:
                errors.append(ann)
            else:
                success.append(ann)

        return success, errors

    def save_to_file(self, save_to, to_format, annotations, item=None):
        """
        Save annotations to a file

        :param save_to:
        :param to_format:
        :param annotations:
        :param item:
        :return:
        """
        # what file format
        if self.save_to_format is None:
            if to_format.lower() in [AnnotationFormat.DATALOOP, AnnotationFormat.COCO]:
                self.save_to_format = 'json'
            elif to_format.lower() in [AnnotationFormat.YOLO]:
                self.save_to_format = 'txt'
            else:
                self.save_to_format = 'xml'

        # save
        # JSON #
        if self.save_to_format == 'json':
            # save json
            save_to = save_to + '.json'
            with open(save_to, "w") as f:
                json.dump(annotations, f, indent=2)

        # TXT #
        elif self.save_to_format == 'txt':
            # save txt
            save_to = save_to + '.txt'
            with open(save_to, "w") as f:
                for ann in annotations:
                    if ann is not None:
                        f.write(' '.join([str(x) for x in ann]) + '\n')

        # XML #
        elif self.save_to_format == 'xml':
            output_annotation = {
                'path': item.filename,
                'filename': os.path.basename(item.filename),
                'folder': os.path.basename(os.path.dirname(item.filename)),
                'width': item.width,
                'height': item.height,
                'depth': 3,
                'database': 'Unknown',
                'segmented': 0,
                'objects': annotations
            }
            save_to = save_to + '.xml'
            environment = Environment(loader=PackageLoader('dtlpy', 'assets'),
                                      keep_trailing_newline=True)
            annotation_template = environment.get_template(self.xml_template_path)
            with open(save_to, 'w') as file:
                content = annotation_template.render(**output_annotation)
                file.write(content)
        else:
            raise exceptions.PlatformException('400', 'Unknown file format to save to')

    def _check_formats(self, from_format, to_format, conversion_func):
        if from_format.lower() not in self.known_formats and conversion_func is None:
            raise exceptions.PlatformException(
                "400",
                "Unknown annotation format: {}, possible formats: {}".format(
                    from_format, self.known_formats
                ),
            )
        if to_format.lower() not in self.known_formats and conversion_func is None:
            raise exceptions.PlatformException(
                "400",
                "Unknown annotation format: {}, possible formats: {}".format(
                    to_format, self.known_formats
                ),
            )

    def convert(self, annotations, from_format, to_format, conversion_func=None, item=None):
        """
        Convert annotations list or single annotation

        :param item:
        :param annotations:
        :param from_format:
        :param to_format:
        :param conversion_func:
        :return:
        """
        # check known format
        self._check_formats(from_format=from_format, to_format=to_format, conversion_func=conversion_func)

        # check annotations param type
        if isinstance(annotations, entities.AnnotationCollection):
            annotations = annotations.annotations
        if not isinstance(annotations, list):
            annotations = [annotations]

        if AnnotationFormat.DATALOOP not in [to_format, from_format]:
            raise exceptions.PlatformException(
                "400", "Can only convert from or to dataloop format"
            )

        # call method
        if conversion_func is None:
            if from_format == AnnotationFormat.DATALOOP:
                method = self.converter_dict[to_format]["to"]
            else:
                method = self.converter_dict[from_format]["from"]
        else:
            method = self.custom_format

        # run all annotations
        pool = ThreadPool(processes=6)
        for i_annotation, annotation in enumerate(annotations):
            pool.apply_async(
                func=self._convert_single,
                kwds={
                    "annotation": annotation,
                    "i_annotation": i_annotation,
                    "conversion_func": conversion_func,
                    'annotations': annotations,
                    'from_format': AnnotationFormat.DATALOOP,
                    'item': item,
                    'method': method
                }
            )

        pool.close()
        pool.join()
        pool.terminate()
        return annotations

    @staticmethod
    def _convert_single(method, **kwargs):
        annotations = kwargs.get('annotations', None)
        i_annotation = kwargs.get('i_annotation', None)
        try:
            ann = method(**kwargs)
            if annotations is not None and isinstance(i_annotation, int):
                annotations[i_annotation] = ann
        except Exception:
            if annotations is not None and isinstance(i_annotation, int):
                annotations[i_annotation] = 'Error converting annotation: {}'.format(traceback.format_exc())
            else:
                raise

    @staticmethod
    def from_voc(annotation, **_):
        bndbox = annotation.find('bndbox')

        if bndbox is None:
            raise Exception('No bndbox field found in annotation object')

        bottom = float(bndbox.find('ymax').text)
        top = float(bndbox.find('ymin').text)
        left = float(bndbox.find('xmin').text)
        right = float(bndbox.find('xmax').text)
        label = annotation.find('name').text

        if annotation.find('segmented'):
            if annotation.find('segmented') == '1':
                logger.warning('Only bounding box conversion is supported in voc format. Segmentation will be ignored.')

        attributes = list(annotation)
        attrs = list()
        for attribute in attributes:
            if attribute.tag not in ['bndbox', 'name'] and len(list(attribute)) == 0:
                if attribute.text not in ['0', '1']:
                    attrs.append(attribute.text)
                elif attribute.text == '1':
                    attrs.append(attribute.tag)

        ann_def = entities.Box(label=label, top=top, bottom=bottom, left=left, right=right, attributes=attrs)
        return entities.Annotation.new(annotation_definition=ann_def)

    def from_yolo(self, annotation, **kwargs):
        (label_id, x, y, w, h) = annotation
        label_id = int(label_id)
        height = kwargs.get('height')
        width = kwargs.get('width')
        item = kwargs.get('item', None)
        if height is None or width is None:
            if item is None:
                raise Exception('Need item width and height in order to convert yolo annotation to dataloop')
            height = item.height
            width = item.width

        x_center = x * width
        y_center = y * height
        w = w * width
        h = h * height

        top = y_center - (h / 2)
        bottom = y_center + (h / 2)
        left = x_center - (w / 2)
        right = x_center + (w / 2)

        label = self._label_by_category_id(category_id=label_id)

        ann_def = entities.Box(label=label, top=top, bottom=bottom, left=left, right=right)
        return entities.Annotation.new(annotation_definition=ann_def, item=item)

    def to_yolo(self, annotation, item=None, **_):
        if not isinstance(annotation, entities.Annotation):
            if item is None:
                item = self.dataset.items.get(item_id=annotation['itemId'])
            annotation = entities.Annotation.from_json(_json=annotation, item=item)
        elif item is None:
            item = annotation.item

        if annotation.type != "box":
            raise Exception('Only box annotations can be converted')
        if item.width is None or item.height is None:
            raise Exception('Item does not have width and height, cannot convert to yolo')

        dw = 1.0 / item.width
        dh = 1.0 / item.height
        x = (annotation.left + annotation.right) / 2.0
        y = (annotation.top + annotation.bottom) / 2.0
        w = annotation.right - annotation.left
        h = annotation.bottom - annotation.top
        x = x * dw
        w = w * dw
        y = y * dh
        h = h * dh

        label_id = self.labels[annotation.label]

        ann = (label_id, x, y, w, h)
        return ann

    def _label_by_category_id(self, category_id):
        if len(self.labels) > 0:
            for label_name, label_index in self.labels.items():
                if label_index == category_id:
                    return label_name
        raise Exception('label category id not found: {}'.format(category_id))

    def from_coco(self, annotation, **kwargs):
        item = kwargs.get('item', None)
        _id = annotation.get('id', None)
        category_id = annotation.get('category_id', None)
        segmentation = annotation.get('segmentation', None)
        iscrowd = annotation.get('iscrowd', None)
        label = self._label_by_category_id(category_id=category_id)

        if hasattr(self, '_only_bbox') and self._only_bbox:
            bbox = annotation.get('bbox', None)
            left = bbox[0]
            top = bbox[1]
            right = left + bbox[2]
            bottom = top + bbox[3]

            ann_def = entities.Box(top=top, left=left, bottom=bottom, right=right, label=label)
        else:
            if int(iscrowd) == 1:
                ann_def = entities.Segmentation(label=label, geo=COCOUtils.rle_to_binary_mask(rle=segmentation))
            else:
                if len(segmentation) == 1:
                    segmentation = segmentation[0]
                    if segmentation:
                        ann_def = entities.Polygon(label=label, is_open=False,
                                                   geo=COCOUtils.rle_to_binary_polygon(segmentation=segmentation))
                    else:
                        bbox = annotation.get('bbox', None)
                        if bbox:
                            left = bbox[0]
                            top = bbox[1]
                            right = left + bbox[2]
                            bottom = top + bbox[3]
                            ann_def = entities.Box(top=top, left=left, bottom=bottom, right=right, label=label)
                        else:
                            raise Exception('Unable to convert annotation, not coordinates')

                else:
                    # TODO - support conversion of split annotations
                    raise exceptions.PlatformException('400',
                                                       'unable to convert.'
                                                       'Converter does not support split annotations: {}'.format(
                                                           _id))

        return entities.Annotation.new(annotation_definition=ann_def, item=item)

    @staticmethod
    def to_coco(annotation, item=None, **kwargs):
        if not isinstance(annotation, entities.Annotation):
            annotation = entities.Annotation.from_json(annotation, item=item)
        area = 0
        iscrowd = 0
        height = kwargs.get('height', item.height)
        width = kwargs.get('width', item.width)

        if annotation.type in ['binary', 'segment']:
            if height is None or width is None:
                raise Exception('Item must have height and width to convert to coco')

        # build annotation
        if annotation.type in ['binary', 'box', 'segment']:
            x = annotation.left
            y = annotation.top
            w = annotation.right - x
            h = annotation.bottom - y
            if annotation.type == 'binary':
                segmentation = COCOUtils.binary_mask_to_rle(binary_mask=annotation.geo, height=height, width=width)
                area = int(annotation.geo.sum())
                iscrowd = 1
            elif annotation.type == 'box':
                segmentation = [[]]
            elif annotation.type == 'segment':
                segmentation, area = COCOUtils.polygon_to_rle(geo=annotation.geo, height=height, width=width)
        else:
            logger.error('Unable to convert annotation of type {} to coco'.format(annotation.type))
            raise Exception('Unable to convert annotation of type {} to coco'.format(annotation.type))

        ann = dict()
        ann['bbox'] = [float(x), float(y), float(w), float(h)]
        ann["segmentation"] = segmentation
        ann["area"] = area
        ann["iscrowd"] = iscrowd

        return ann

    @staticmethod
    def to_voc(annotation, item=None, **_):
        if not isinstance(annotation, entities.Annotation):
            annotation = entities.Annotation.from_json(annotation, item=item)

        if annotation.type != 'box':
            raise exceptions.PlatformException('400', 'Annotation must be of type box')

        label = annotation.label

        if annotation.attributes is not None and isinstance(annotation.attributes, list) and len(
                annotation.attributes) > 0:
            attributes = annotation.attributes
        else:
            attributes = list()

        left = annotation.left
        top = annotation.top
        right = annotation.right
        bottom = annotation.bottom

        ann = {'name': label,
               'xmin': left,
               'ymin': top,
               'xmax': right,
               'ymax': bottom,
               'attributes': attributes,
               }

        return ann

    @staticmethod
    def custom_format(annotation, conversion_func, i_annotation=None, annotations=None, from_format=None,
                      item=None, **_):
        if from_format == AnnotationFormat.DATALOOP and isinstance(annotation, dict):
            annotation = entities.Annotation.from_json(_json=annotation, item=item)

        ann = conversion_func(annotation)

        if isinstance(annotations[i_annotation], entities.Annotation) and annotations[i_annotation].item is None:
            ann.item = item

        return ann



import numpy as np

from . import Box


class Note(Box):
    """
        Note annotation object
    """

    def __init__(self, left, top, right, bottom, label, attributes=None, messages=None, status=None,
                 create_time=None, creator=None):
        super(Note, self).__init__(left=left, top=top, right=right, bottom=bottom, label=label, attributes=attributes)
        self.type = "note"
        self.messages = messages
        self.status = status
        self.create_time = create_time
        self.creator = creator

    def to_coordinates(self, color):
        box = super(Note, self).to_coordinates(color=color)
        note = {
            'messages': self.messages,
            'status': self.status,
            'createTime': self.create_time,
            'creator': self.creator
        }
        coordinates = {
            'box': box,
            'note': note
        }

        return coordinates

    @staticmethod
    def from_coordinates(coordinates):
        return Box.from_coordinates(coordinates['box'])

    @classmethod
    def from_json(cls, _json):
        if "coordinates" in _json:
            geo = cls.from_coordinates(_json["coordinates"])
            note_data = _json["coordinates"].get('note', dict())
        elif "data" in _json:
            geo = cls.from_coordinates(_json["data"])
            note_data = _json["data"].get('note', dict())
        else:
            raise ValueError('can not find "coordinates" or "data" in annotation. id: {}'.format(_json["id"]))

        left = np.min(geo[:, 0])
        top = np.min(geo[:, 1])
        right = np.max(geo[:, 0])
        bottom = np.max(geo[:, 1])

        attributes = _json.get("attributes", list())
        messages = note_data.get('messages', list())

        return cls(
            left=left,
            top=top,
            right=right,
            bottom=bottom,
            label=_json["label"],
            attributes=attributes,
            messages=messages,
            status=note_data.get('status', 'open'),
            creator=note_data.get('creator', 'me'),
            create_time=note_data.get('createTime', 0),
        )



from collections import namedtuple
import traceback
import logging
import attr

from .. import repositories, entities, services

logger = logging.getLogger(name=__name__)


@attr.s
class Snapshot(entities.BaseEntity):
    """
    Snapshot object
    """
    # platform
    id = attr.ib()
    project_id = attr.ib()
    dataset_id = attr.ib(repr=False)
    model_id = attr.ib(repr=False)
    parent_snapshot_id = attr.ib(repr=False)
    artifacts_id = attr.ib()

    name = attr.ib()
    version = attr.ib()
    query = attr.ib(repr=False)

    url = attr.ib(repr=False)
    createdAt = attr.ib()
    updatedAt = attr.ib(repr=False)
    creator = attr.ib()

    # sdk
    _model = attr.ib(repr=False)
    _project = attr.ib(repr=False)
    _client_api = attr.ib(type=services.ApiClient, repr=False)
    _repositories = attr.ib(repr=False)

    @staticmethod
    def _protected_from_json(_json, client_api, project, model, is_fetched=True):
        """
        Same as from_json but with try-except to catch if error

        :param _json: platform representation of snapshot
        :param client_api: ApiClient entity
        :param project: project that owns the snapshot
        :param model: model entity of the snapshot
        :param is_fetched: is Entity fetched from Platform
        :return: Snapshot entity
        """
        try:
            snapshot = Snapshot.from_json(_json=_json,
                                          client_api=client_api,
                                          project=project,
                                          model=model,
                                          is_fetched=is_fetched)
            status = True
        except Exception:
            snapshot = traceback.format_exc()
            status = False
        return status, snapshot

    @classmethod
    def from_json(cls, _json, client_api, project, model, is_fetched=True):
        """
        Turn platform representation of snapshot into a snapshot entity

        :param _json: platform representation of snapshot
        :param client_api: ApiClient entity
        :param project: project that owns the snapshot
        :param model: model entity of the snapshot
        :param is_fetched: is Entity fetched from Platform
        :return: Snapshot entity
        """
        inst = cls(
            project_id=_json.get('projectId', None),
            dataset_id=_json.get('datasetId', None),
            model_id=_json.get('modelId', None),
            parent_snapshot_id=_json.get('parentCheckpointId', None),
            artifacts_id=_json.get('artifactId', None),
            query=_json.get('query', None),
            createdAt=_json.get('createdAt', None),
            updatedAt=_json.get('updatedAt', None),
            version=_json.get('version', None),
            creator=_json.get('creator', None),
            client_api=client_api,
            name=_json.get('name', None),
            url=_json.get('url', None),
            project=project,
            model=model,
            id=_json.get('id', None)
        )
        inst.is_fetched = is_fetched
        return inst

    def to_json(self):
        """
        Turn Snapshot entity into a platform representation of Snapshot

        :return: platform json of snapshot
        """
        _json = attr.asdict(self,
                            filter=attr.filters.exclude(attr.fields(Snapshot)._project,
                                                        attr.fields(Snapshot)._model,
                                                        attr.fields(Snapshot)._repositories,
                                                        attr.fields(Snapshot)._client_api,
                                                        attr.fields(Snapshot).model_id,
                                                        attr.fields(Snapshot).parent_snapshot_id,
                                                        attr.fields(Snapshot).project_id,
                                                        attr.fields(Snapshot).dataset_id,
                                                        attr.fields(Snapshot).artifacts_id,
                                                        ))

        _json['projectId'] = self.project_id
        _json['datasetId'] = self.dataset_id
        _json['parentCheckpointId'] = self.parent_snapshot_id
        _json['artifactId'] = self.artifacts_id
        _json['modelId'] = self.model_id
        return _json

    ############
    # entities #
    ############
    @property
    def project(self):
        if self._project is None:
            self._project = self.projects.get(project_id=self.project_id, fetch=None)
        assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def model(self):
        if self._model is None:
            self._model = self.models.get(model_id=self.model_id)
        assert isinstance(self._model, entities.Model)
        return self._model

    ################
    # repositories #
    ################
    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['projects', 'snapshots', 'models'])

        r = reps(projects=repositories.Projects(client_api=self._client_api),
                 snapshots=repositories.Snapshots(client_api=self._client_api,
                                                  project=self._project,
                                                  project_id=self.project_id,
                                                  model=self.model),
                 models=repositories.Models(client_api=self._client_api, project=self._project))
        return r

    @property
    def projects(self):
        assert isinstance(self._repositories.projects, repositories.Projects)
        return self._repositories.projects

    @property
    def snapshots(self):
        assert isinstance(self._repositories.snapshots, repositories.Snapshots)
        return self._repositories.snapshots

    @property
    def models(self):
        assert isinstance(self._repositories.models, repositories.Models)
        return self._repositories.models

    ###########
    # methods #
    ###########
    def update(self):
        """
        Update Snapshot changes to platform

        :return: Snapshot entity
        """
        return self.snapshots.update(snapshot=self)

    def upload(self, local_path, overwrite=True):
        """
        Upload files to existing Snapshot

        :return: Snapshot entity
        """
        return self.snapshots.upload(snapshot=self, local_path=local_path, overwrite=overwrite)

    def download(self, local_path):
        """
        Download artifacts from snapshot
        """
        return self.snapshots.download(snapshot=self, local_path=local_path)

    def delete(self):
        """
        Delete Snapshot object

        :return: True
        """
        return self.snapshots.delete(snapshot=self)



import logging
import numpy as np

logger = logging.getLogger(name=__name__)


class BaseAnnotationDefinition:
    @staticmethod
    def add_text_to_image(image, annotation):
        try:
            import cv2
        except ImportError:
            logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise

        text = '{label}-{attributes}'.format(label=annotation.label, attributes=','.join(annotation.attributes))
        top = annotation.top
        left = annotation.left
        if top == 0:
            top = image.shape[0] / 10
        if left == 0:
            left = image.shape[1] / 10
        return cv2.putText(img=image,
                           text=text,
                           org=tuple([int(np.round(top)), int(np.round(left))]),
                           color=(255, 0, 0),
                           fontFace=cv2.FONT_HERSHEY_DUPLEX,
                           fontScale=1,
                           thickness=2)

    @property
    def logger(self):
        return logger



from collections import namedtuple
import traceback
import logging
import attr

from .. import repositories, entities, services

logger = logging.getLogger(name=__name__)


@attr.s
class Model(entities.BaseEntity):
    """
    Model object
    """
    # platform
    id = attr.ib()
    url = attr.ib(repr=False)
    version = attr.ib()
    createdAt = attr.ib()
    updatedAt = attr.ib(repr=False)
    name = attr.ib()
    description = attr.ib()
    codebase_id = attr.ib()
    entry_point = attr.ib()
    creator = attr.ib()

    # name change
    project_id = attr.ib()

    # sdk
    _project = attr.ib(repr=False)
    _client_api = attr.ib(type=services.ApiClient, repr=False)
    _repositories = attr.ib(repr=False)

    @staticmethod
    def _protected_from_json(_json, client_api, project, is_fetched=True):
        """
        Same as from_json but with try-except to catch if error
        :param _json: platform representation of model
        :param client_api:
        :param project:
        :param is_fetched: is Entity fetched from Platform
        :return:
        """
        try:
            model = Model.from_json(_json=_json,
                                    client_api=client_api,
                                    project=project,
                                    is_fetched=is_fetched)
            status = True
        except Exception:
            model = traceback.format_exc()
            status = False
        return status, model

    @classmethod
    def from_json(cls, _json, client_api, project, is_fetched=True):
        """
        Turn platform representation of model into a model entity

        :param _json: platform representation of model
        :param client_api:
        :param project:
        :param is_fetched: is Entity fetched from Platform
        :return: Model entity
        """
        inst = cls(
            project_id=_json.get('projectId', None),
            codebase_id=_json.get('codebaseId', None),
            createdAt=_json.get('createdAt', None),
            updatedAt=_json.get('updatedAt', None),
            version=_json.get('version', None),
            description=_json.get('description', None),
            creator=_json.get('creator', None),
            entry_point=_json.get('entryPoint', None),
            client_api=client_api,
            name=_json.get('name', None),
            url=_json.get('url', None),
            project=project,
            id=_json.get('id', None)
        )
        inst.is_fetched = is_fetched
        return inst

    def to_json(self):
        """
        Turn Model entity into a platform representation of Model

        :return: platform json of model
        """
        _json = attr.asdict(self,
                            filter=attr.filters.exclude(attr.fields(Model)._project,
                                                        attr.fields(Model)._repositories,
                                                        attr.fields(Model)._client_api,
                                                        attr.fields(Model).codebase_id,
                                                        attr.fields(Model).entry_point,
                                                        attr.fields(Model).project_id,
                                                        ))

        _json['projectId'] = self.project_id
        _json['entryPoint'] = self.entry_point
        _json['codebaseId'] = self.codebase_id
        return _json

    ############
    # entities #
    ############
    @property
    def project(self):
        if self._project is None:
            self._project = self.projects.get(project_id=self.project_id, fetch=None)
        assert isinstance(self._project, entities.Project)
        return self._project

    ################
    # repositories #
    ################
    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['projects', 'models', 'snapshots'])

        r = reps(projects=repositories.Projects(client_api=self._client_api),
                 models=repositories.Models(client_api=self._client_api, project=self._project),
                 snapshots=repositories.Snapshots(client_api=self._client_api,
                                                  project=self._project,
                                                  model=self),
                 )
        return r

    @property
    def projects(self):
        assert isinstance(self._repositories.projects, repositories.Projects)
        return self._repositories.projects

    @property
    def snapshots(self):
        assert isinstance(self._repositories.snapshots, repositories.Snapshots)
        return self._repositories.snapshots

    @property
    def models(self):
        assert isinstance(self._repositories.models, repositories.Models)
        return self._repositories.models

    ##############
    # properties #
    ##############
    @property
    def git_status(self):
        status = 'Git status unavailable'
        try:
            codebase = self.project.codebases.get(codebase_id=self.codebase_id, version=self.version - 1)
            if 'git' in codebase.metadata:
                status = codebase.metadata['git'].get('status', status)
        except Exception:
            logging.debug('Error getting codebase')
        return status

    @property
    def git_log(self):
        log = 'Git log unavailable'
        try:
            codebase = self.project.codebases.get(codebase_id=self.codebase_id, version=self.version - 1)
            if 'git' in codebase.metadata:
                log = codebase.metadata['git'].get('log', log)
        except Exception:
            logging.debug('Error getting codebase')
        return log

    ###########
    # methods #
    ###########
    def update(self):
        """
        Update Model changes to platform

        :return: Model entity
        """
        return self.models.update(model=self)

    def checkout(self):
        """
        Checkout as model

        :return:
        """
        return self.models.checkout(model=self)

    def delete(self):
        """
        Delete Model object

        :return: True
        """
        return self.models.delete(model=self)

    def push(self, codebase_id=None, src_path=None, model_name=None, modules=None, checkout=False):
        """
         Push local model

        :param checkout:
        :param codebase_id:
        :param src_path:
        :param model_name:
        :param modules:
        :return:
        """
        return self.project.models.push(model_name=model_name if model_name is not None else self.name,
                                        codebase_id=codebase_id,
                                        src_path=src_path,
                                        modules=modules,
                                        checkout=checkout)

    def build(self, local_path=None, from_local=None):
        """
        Push local model

        :param from_local: bool. use current directory to build
        :param local_path: local path of the model (if from_local=False - codebase will be downloaded)
        :return:
        """
        return self.models.build(model=self,
                                 local_path=local_path,
                                 from_local=from_local)



#! /usr/bin/env python3
# This file is part of DTLPY.
#
# DTLPY is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DTLPY is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with DTLPY.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup, find_packages

with open('README.md') as f:
    readme = f.read()

with open('LICENSE') as f:
    license_ = f.read()

with open('requirements.txt') as f:
    requirements = f.read()

setup(name='dtlpy',
      version='1.17.12',
      description='SDK and CLI for Dataloop platform',
      author='Or Shabtay',
      author_email='or@dataloop.ai',
      url='https://github.com/dataloop-ai/dtlpy',
      license='Apache License 2.0',
      long_description=readme,
      long_description_content_type='text/markdown',
      packages=find_packages(exclude=('tests', 'docs', 'samples')),
      setup_requires=['wheel'],
      install_requires=requirements,
      test_suite='tests',
      python_requires='>=3.5.4',
      scripts=['dtlpy/dlp/dlp.py', 'dtlpy/dlp/dlp.bat', 'dtlpy/dlp/dlp'],
      include_package_data=True,
      entry_points={
          'console_scripts': [
              'dlp=dtlpy.dlp.dlp:main',
          ],
      },
      )



import logging
import uuid
import os

from .. import entities, repositories, exceptions, miscellaneous, services

logger = logging.getLogger(name=__name__)


class Snapshots:
    """
    Snapshots Repository
    """

    def __init__(self,
                 client_api: services.ApiClient,
                 model: entities.Model = None,
                 project: entities.Project = None,
                 artifacts: repositories.Artifacts = None,
                 project_id: str = None):
        self._client_api = client_api
        self._project = project
        self._model = model
        self._artifacts = artifacts
        self._project_id = project_id

    ############
    # entities #
    ############
    @property
    def project(self) -> entities.Project:
        if self._project is None:
            if self._project_id is not None:
                projects = repositories.Projects(client_api=self._client_api)
                self._project = projects.get(project_id=self._project_id)
        if self._project is None:
            if self._model is not None:
                if self._model._project is not None:
                    self._project = self._model._project
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "project". need to set a Project entity or use project.snapshots repository')
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    @property
    def artifacts(self) -> repositories.Artifacts:
        assert isinstance(self.project.artifacts, repositories.Artifacts)
        return self.project.artifacts

    @property
    def model(self) -> entities.Model:
        if self._model is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Cannot perform action WITHOUT Model entity in Datasets repository.'
                        ' Please use model.snapshots or set a model')
        assert isinstance(self._model, entities.Model)
        return self._model
    ###########
    # methods #
    ###########
    def get(self, snapshot_name=None, snapshot_id=None) -> entities.Snapshot:
        """
        Get snapshot object

        :param snapshot_id:
        :param snapshot_name:
        :return: snapshot object
        """

        if snapshot_id is not None:
            success, response = self._client_api.gen_request(req_type="get",
                                                             path="/checkpoints/{}".format(snapshot_id))
            if not success:
                raise exceptions.PlatformException(response)
            snapshot = entities.Snapshot.from_json(client_api=self._client_api,
                                                   _json=response.json(),
                                                   project=self._project,
                                                   model=self._model)
        elif snapshot_name is not None:
            snapshots = self.list(entities.Filters(resource=entities.FiltersResource.SNAPSHOT,
                                                   field='name',
                                                   values=snapshot_name))
            if snapshots.items_count == 0:
                raise exceptions.PlatformException(
                    error='404',
                    message='Snapshot not found. Name: {}'.format(snapshot_name))
            elif snapshots.items_count > 1:
                raise exceptions.PlatformException(
                    error='400',
                    message='More than one file found by the name of: {}'.format(snapshot_name))
            snapshot = snapshots[0]
        else:
            raise exceptions.PlatformException(
                error='400',
                message='Checked out not found, must provide either snapshot id or snapshot name')

        return snapshot

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Snapshot]:
        jobs = [None for _ in range(len(response_items))]
        pool = self._client_api.thread_pools(pool_name='entity.create')

        # return triggers list
        for i_service, service in enumerate(response_items):
            jobs[i_service] = pool.apply_async(entities.Snapshot._protected_from_json,
                                               kwds={'client_api': self._client_api,
                                                     '_json': service,
                                                     'model': self._model,
                                                     'project': self._project})
        # wait for all jobs
        _ = [j.wait() for j in jobs]
        # get all results
        results = [j.get() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        return miscellaneous.List([r[1] for r in results if r[0] is True])

    def _list(self, filters: entities.Filters):
        url = '/query/machine-learning'
        # request
        success, response = self._client_api.gen_request(req_type='POST',
                                                         path=url,
                                                         json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    def list(self, filters: entities.Filters = None) -> entities.PagedEntities:
        """
        List project snapshots
        :return:
        """
        # default filters
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.SNAPSHOT)
            if self._project is not None:
                filters.add(field='projectId', values=self._project.id)
            if self._model is not None:
                filters.add(field='modelId', values=self._model.id)

        # assert type filters
        if not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException('400', 'Unknown filters type')

        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       client_api=self._client_api)
        paged.get_page()
        return paged

    def create(self, snapshot_name=None, description=None, project_id=None, scope='private') -> entities.Snapshot:
        """
            Create a Snapshot entity

        :param snapshot_name: name of the snapshot
        :param description: description
        :param project_id: project that owns the snapshot
        :param scope: 'global'
        :return: Snapshot Entity
        """
        if project_id is None:
            project_id = self.project.id
        else:
            if self._project is not None and self._project.id != project_id:
                self._project = None
            self._project_id = project_id
        artifact_dir_item = self.artifacts.items_repository.make_dir(
            directory=self.artifacts._build_path_header(model_name=self.model.name,
                                                        snapshot_name='{}_{}'.format(snapshot_name,
                                                                                     str(uuid.uuid1())),
                                                        ))
        # create payload for request
        payload = {'name': snapshot_name,
                   'description': description,
                   'modelId': self.model.id,
                   'projectId': project_id,
                   'artifactId': artifact_dir_item.id,
                   'scope': scope}

        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/checkpoints',
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        snapshot = entities.Snapshot.from_json(_json=response.json(),
                                               client_api=self._client_api,
                                               project=self._project,
                                               model=self.model)
        return snapshot

    def upload(self,
               local_path: str,
               snapshot: entities.Snapshot = None,
               overwrite: bool = False):
        """
        Create a snapshot in platform

        :param local_path: path of artifacts to upload
        :param snapshot: Snapshot entity
        :param overwrite: overwrite the artifacts (if same name exists)
        :return: Snapshot Entity
        """

        # upload artifacts
        if not local_path.endswith('/'):
            local_path += '/'
        # snapshot binaries must be flatten (no directory trees)
        local_path += '*'
        artifact_dir_item = self.artifacts.items_repository.get(item_id=snapshot.artifacts_id)
        artifacts = self.artifacts.items_repository.upload(local_path=local_path,
                                                           remote_path=artifact_dir_item.filename,
                                                           overwrite=overwrite)
        return artifacts

    def download(self, snapshot_id=None, snapshot=None, local_path=None):
        if snapshot is None:
            if snapshot_id is None:
                raise exceptions.PlatformException('400', 'Please provide snapshot or snapshot id')
            snapshot = self.get(snapshot_id=snapshot_id)

        if local_path is None:
            local_path = os.getcwd()

        snapshot_artifact_item = self.artifacts.get(artifact_id=snapshot.artifacts_id)
        if snapshot_artifact_item.type == 'dir':
            filters = entities.Filters(field='dir', values='{}*'.format(snapshot_artifact_item.filename))
            artifacts = self.artifacts.items_repository.download(
                filters=filters,
                local_path=local_path,
                to_items_folder=False,
                without_relative_path=snapshot_artifact_item.filename)
        else:
            artifacts = snapshot_artifact_item.download(local_path=local_path)
        return artifacts

    def delete(self, snapshot: entities.Snapshot = None, snapshot_name=None, snapshot_id=None):
        """
        Delete Snapshot object

        :param snapshot:
        :param snapshot_name:
        :param snapshot_id:
        :return: True
        """
        # get id and name
        if snapshot is None:
            snapshot = self.get(snapshot_id=snapshot_id, snapshot_name=snapshot_name)
        try:
            self.artifacts.items_repository.delete(item_id=snapshot.artifacts_id)
        except exceptions.NotFound:
            pass

        # request
        success, response = self._client_api.gen_request(
            req_type="delete",
            path="/checkpoints/{}".format(snapshot.id)
        )

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return results
        return True

    def update(self, snapshot: entities.Snapshot) -> entities.Snapshot:
        """
        Update Snapshot changes to platform

        :param snapshot:
        :return: Snapshot entity
        """
        # payload
        payload = snapshot.to_json()

        # request
        success, response = self._client_api.gen_request(req_type='patch',
                                                         path='/checkpoints/{}'.format(snapshot.id),
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.Snapshot.from_json(_json=response.json(),
                                           client_api=self._client_api,
                                           project=self._project,
                                           model=snapshot._model)



import threading


class BaseServiceRunner:
    _do_reset = False
    _auto_refresh_dtlpy_token = None
    _refresh_dtlpy_token = None
    _threads_terminated = list()
    _threads_terminated_lock = threading.Lock()

    def do_reset(self):
        self._do_reset = True

    def _terminate(self, tid):
        with self._threads_terminated_lock:
            self._threads_terminated.append(tid)

    def kill_event(self):
        ident = threading.get_ident()
        if ident in self._threads_terminated:
            with self._threads_terminated_lock:
                self._threads_terminated.pop(self._threads_terminated.index(ident))
            raise InterruptedError('Execution received termination signal')


class Progress:

    def update(self, status=None, progress=0, message=None, output=None):
        pass



import json
import io
from copy import deepcopy
from .. import entities, exceptions


class CollectionTypes:
    SIMILARITY = 'collection'
    MULTIVIEW = 'multi'


class SimilarityTypeEnum:
    """
    State enum
    """
    ID = "id"
    URL = "url"


class CollectionItem:
    """
    Base CollectionItem
    """

    def __init__(self, type, ref):
        assert isinstance(ref, str)
        self.ref = ref
        self.type = type

    @classmethod
    def from_json(cls, _json):
        return cls(
            type=_json['type'],
            ref=_json['ref']
        )

    def to_json(self):
        _json = {
            'type': self.type,
            'ref': self.ref
        }

        return _json


class SimilarityItem(CollectionItem):
    """
    Single similarity item
    """

    def __init__(self, type, ref, target=False):
        super().__init__(type=type, ref=ref)
        assert isinstance(ref, str)
        self.target = target

    def to_json(self):
        _json = super().to_json()

        if self.target:
            _json['target'] = self.target

        return _json


class MultiViewItem(CollectionItem):
    """
    Single multi view item
    """

    def __init__(self, type, ref):
        super().__init__(type=type, ref=ref)


class Collection:
    """
    Base Collection Entity
    """

    def __init__(self, type, name, items=None):
        self.type = type
        self.name = name
        self._items = self._items_to_list(items=items)

    @staticmethod
    def _items_to_list(items):
        items_list = list()
        if items:
            if not isinstance(items, list):
                items = [items]

            for item in items:
                if isinstance(item, entities.Item):
                    items_list.append({'type': SimilarityTypeEnum.ID, 'ref': item.id})
                elif isinstance(item, SimilarityItem) or isinstance(item, MultiViewItem):
                    items_list.append(item.to_json())
                elif isinstance(item, dict):
                    items_list.append(item)
                else:
                    raise exceptions.PlatformException('400',
                                                       'items must be a list of the following: '
                                                       'Item, MultiViewItem, SimilarityItem, dict')
        return items_list

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        """
        _json = {
            "type": self.type,
            "shebang": "dataloop",
            "metadata": {
                "dltype": self.type,
            }
        }

        return _json

    def to_bytes_io(self):
        byte_io = io.BytesIO()
        byte_io.name = self.name
        byte_io.write(json.dumps(self.to_json()).encode())
        byte_io.seek(0)

        return byte_io

    def add(self, ref, type=SimilarityTypeEnum.ID):
        """
        Add item to collection
        """
        item = {
            'ref': ref,
            'type': type
        }

        self._items.append(item)

    def pop(self, ref):
        for item in self._items:
            if item['ref'] == ref:
                self._items.remove(item)


class Similarity(Collection):
    """
    Similarity Entity
    """

    def __init__(self, ref, name=None, items=None):
        if name is None:
            name = ref
        super().__init__(type=CollectionTypes.SIMILARITY, name=name, items=items)
        assert isinstance(ref, str)
        self.ref = ref

    @property
    def items(self):
        """
        list of the collection items
        """
        return [SimilarityItem.from_json(_json=item) for item in self._items]

    @property
    def target(self):
        """
        Target item for similiraty
        """
        return SimilarityItem(ref=self.ref, type=self.type, target=True)

    @classmethod
    def from_json(cls, _json):
        return cls(
            ref=_json['metadata']['target']['ref'],
            items=_json.get('items', list())
        )

    def _fixed_items(self):
        items = deepcopy(self._items)
        target_in_items = False
        for item in items:
            if item.get('ref', '') == self.ref:
                target_in_items = True
                item['target'] = True
            else:
                item.pop('target', None)

        if not target_in_items:
            items.append(self.target.to_json())

        return items

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        """
        _json = super().to_json()
        _json['metadata']['target'] = {
            "type": self.type,
            "ref": self.ref
        }
        _json['items'] = self._fixed_items()

        return _json


class MultiView(Collection):
    """
    Multi Entity
    """

    def __init__(self, name, items=None):
        super().__init__(type=CollectionTypes.MULTIVIEW, name=name, items=items)

    @property
    def items(self):
        """
        list of the collection items
        """
        return [MultiViewItem.from_json(_json=item) for item in self._items]

    @classmethod
    def from_json(cls, _json):
        return cls(items=_json.get('items', list()))

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        """
        _json = super().to_json()
        _json['items'] = self._items

        return _json



import dictdiffer
import logging

logger = logging.getLogger(name=__name__)


# noinspection PyPep8Naming
class DictDiffer:

    @staticmethod
    def diff(origin, modified):
        TYPE = 0
        FIELD = 1
        LIST = 2

        diffs = dict()
        dict_diff = list(dictdiffer.diff(origin, modified))
        for diff in dict_diff:
            field_pointer = DictDiffer.get_field_path(diffs=diffs, path=diff[FIELD], diff_type=diff[TYPE])
            if diff[TYPE] == 'add':
                for addition in diff[LIST]:
                    field_pointer[addition[0]] = addition[1]

            elif diff[TYPE] == 'remove':
                for deletion in diff[LIST]:
                    field_pointer[deletion[0]] = None

            elif diff[TYPE] == 'change':
                change = diff[LIST]
                field = diff[FIELD]
                if not isinstance(field, list):
                    field = field.split('.')
                field_pointer[field[-1]] = change[1]
        return diffs

    @staticmethod
    def get_field_path(diffs, path, diff_type):
        field_pointer = diffs
        if not isinstance(path, list):
            path = path.split('.')

        if len(path) > 1 or diff_type != 'change':
            for level in path:
                if diff_type == 'change' and level == path[-2]:
                    if level not in field_pointer:
                        field_pointer[level] = dict()
                    field_pointer = field_pointer[level]
                    break
                if level in field_pointer:
                    field_pointer = field_pointer[level]
                else:
                    field_pointer[level] = dict()
                    field_pointer = field_pointer[level]

        return field_pointer



try:
    import dtlpy as dl
except ImportError:
    print('Dataloop SDK not found, you can only work locally. "pip install dtlpy" to install')


class AdapterModel:
    def __init__(self, device, model_specs, hp_values, final):
        self.model_name = ''
        self.model = None

    def reformat(self):
        pass

    def data_loader(self):
        pass

    def preprocess(self, batch):
        return batch

    def load(self, local_path):
        raise NotImplementedError

    def load_from_snapshot(self, local_path, snapshot_id):
        snapshot = dl.snapshots.get(snapshot_id=snapshot_id)
        snapshot.download(local_path=local_path)
        self.load(local_path)

    def save(self, local_path):
        self.model.save(local_path)

    def save_to_snapshot(self, local_path):
        self.save(local_path=local_path)
        dl.snapshots.upload()

    def train(self):
        pass

    def get_snapshot(self):
        pass

    def get_metrics(self):
        pass

    def predict_items(self, items, with_upload=True):
        for item in items:
            filepath = item.download()
            results = self.predict(filepath)
            if with_upload:
                builder = item.annotation.builder()
                for res in results:
                    builder.add(dl.Segmaeations())
                item.annotations.upload(results)

    def predict(self, batch):
        raise NotImplementedError



import logging
import traceback

from .. import entities, miscellaneous, exceptions, services

logger = logging.getLogger(name=__name__)


class Ontologies:
    """
    Ontologies repository
    """

    def __init__(self, client_api: services.ApiClient, recipe: entities.Recipe = None):
        self._client_api = client_api
        self._recipe = recipe

    ############
    # entities #
    ############
    @property
    def recipe(self) -> entities.Recipe:
        if self._recipe is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "recipe". need to set a Recipe entity or use ontology.recipes repository')
        assert isinstance(self._recipe, entities.Recipe)
        return self._recipe

    @recipe.setter
    def recipe(self, recipe: entities.Recipe):
        if not isinstance(recipe, entities.Recipe):
            raise ValueError('Must input a valid Recipe entity')
        self._recipe = recipe

    ###########
    # methods #
    ###########
    def create(self, labels, project_ids=None, attributes=None) -> entities.Ontology:
        """
        Create a new ontology

        :param labels: recipe tags
        :param project_ids: recipe project/s
        :param attributes: recipe attributes
        :return: Ontology object
        """
        if attributes is None:
            attributes = list()
        if project_ids is None and self._recipe is not None:
            project_ids = [self.recipe.dataset.project.id]
        elif not isinstance(project_ids, list):
            project_ids = [project_ids]
        # convert to platform label format (root)
        labels = self.labels_to_roots(labels)
        payload = {"roots": labels,
                   "projectIds": project_ids,
                   "attributes": attributes}
        success, response = self._client_api.gen_request(req_type="post",
                                                         path="/ontologies",
                                                         json_req=payload)
        if success:
            logger.info("Ontology was created successfully")
            ontology = entities.Ontology.from_json(_json=response.json(),
                                                   client_api=self._client_api,
                                                   recipe=self._recipe)
        else:
            logger.error("Failed to create Ontology")
            raise exceptions.PlatformException(response)
        return ontology

    def list(self) -> miscellaneous.List[entities.Ontology]:
        """
        List ontologies for recipe

        :return:
        """
        if self._recipe is None:
            raise ("400", "Action is not permitted")

        ontologies = [ontology_id for ontology_id in self.recipe.ontologyIds]

        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(ontologies))]
        for i_ontology, ontology_id in enumerate(ontologies):
            jobs[i_ontology] = pool.apply_async(self._protected_get, kwds={'ontology_id': ontology_id})
        # wait for all jobs
        _ = [j.wait() for j in jobs]
        # get all results
        results = [j.get() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        return miscellaneous.List([r[1] for r in results if r[0] is True])

    def _protected_get(self, ontology_id):
        """
        Same as get but with try-except to catch if error
        :param ontology_id:
        :return:
        """
        try:
            ontology = self.get(ontology_id=ontology_id)
            status = True
        except Exception:
            ontology = traceback.format_exc()
            status = False
        return status, ontology

    def get(self, ontology_id) -> entities.Ontology:
        """
        Get Ontology object

        :param ontology_id: ontology id
        :return: Ontology object
        """
        success, response = self._client_api.gen_request(req_type="get",
                                                         path="/ontologies/{}".format(ontology_id))
        if success:
            ontology = entities.Ontology.from_json(_json=response.json(),
                                                   client_api=self._client_api,
                                                   recipe=self._recipe)
        else:
            raise exceptions.PlatformException(response)
        return ontology

    def delete(self, ontology_id):
        """
        Delete Ontology from platform

        :param ontology_id: ontology_id id
        :return: True
        """
        success, response = self._client_api.gen_request(req_type="delete",
                                                         path="/ontologies/%s" % ontology_id)
        if success:
            logger.debug("Ontology was deleted successfully")
            return success
        else:
            raise exceptions.PlatformException(response)

    def update(self, ontology: entities.Ontology, system_metadata=False) -> entities.Ontology:
        """
        Update Ontology metadata

       :param ontology: Ontology object
       :param system_metadata: bool
       :return: Ontology object
       """
        url_path = "/ontologies/%s" % ontology.id
        if system_metadata:
            url_path += "?system=true"
        success, response = self._client_api.gen_request(req_type="put",
                                                         path=url_path,
                                                         json_req=ontology.to_json())
        if success:
            logger.debug("Ontology was updated successfully")
            # update dataset labels
            ontology = entities.Ontology.from_json(_json=response.json(),
                                                   client_api=self._client_api,
                                                   recipe=self._recipe)
            self.recipe.dataset._labels = ontology.labels
            return ontology
        else:
            logger.error("Failed to update ontology:")
            raise exceptions.PlatformException(response)

    @staticmethod
    def labels_to_roots(labels):
        """
        Converts labels dict to a list of platform representation of labels

        :param labels: labels dict
        :return: platform representation of labels
        """
        roots = list()
        if isinstance(labels, dict):
            for label in labels:
                root = {
                    "value": {
                        "tag": label,
                        "color": labels[label],
                        "attributes": list(),
                    },
                    "children": list(),
                }
                roots.append(root)
        elif isinstance(labels, list):
            for label in labels:
                if isinstance(label, entities.Label):
                    root = label.to_root()
                elif "value" in label:
                    root = {
                        "value": label["value"],
                        "children": label.get("children", list()),
                    }
                else:
                    root = {
                        "value": {
                            "tag": label.get("tag", None),
                            "color": label.get("color", "#FFFFFF"),
                            "attributes": label.get("attributes", list()),
                        },
                        "children": label.get("children", list()),
                    }
                roots.append(root)
        for root in roots:
            if not isinstance(root["value"]["color"], str):
                # noinspection PyStringFormat
                root["value"]["color"] = "#%02x%02x%02x" % root["value"]["color"]
        return roots



import datetime
import inspect
import logging
import json
import os
import tempfile
from typing import Union, List, Callable

from .. import miscellaneous, exceptions, entities, repositories, assets, ApiClient
from ..__version__ import version as __version__

logger = logging.getLogger(name=__name__)


class Services:
    def __init__(self,
                 client_api: ApiClient,
                 project: entities.Project = None,
                 package: entities.Package = None,
                 project_id=None):
        self._client_api = client_api
        self._package = package
        self._project = project
        if project_id is None:
            if project is not None:
                project_id = project.id
        self._project_id = project_id

    ############
    # entities #
    ############
    @property
    def package(self) -> entities.Package:
        if self._package is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Cannot perform action WITHOUT package entity in services repository. Please set a package')
        assert isinstance(self._package, entities.Package)
        return self._package

    @package.setter
    def package(self, package: entities.Package):
        if not isinstance(package, entities.Package):
            raise ValueError('Must input a valid package entity')
        self._package = package

    @property
    def project(self) -> entities.Project:
        if self._project is None:
            # try to get from package
            if self._package is not None:
                self._project = self._package._project

        if self._project is None:
            # try to get checked out project
            project = self._client_api.state_io.get('project')
            if project is not None:
                self._project = entities.Project.from_json(_json=project, client_api=self._client_api)
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Cannot perform action WITHOUT Project entity in services repository. Please set a project')
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    def open_in_web(self, service=None, service_id=None, service_name=None):
        if service is None:
            service = self.get(service_id=service_id, service_name=service_name)
        self._client_api._open_in_web(resource_type='service', project_id=service.project_id,
                                      package_id=service.package_id, service_id=service.id)

    def __get_from_cache(self) -> entities.Service:
        service = self._client_api.state_io.get('service')
        if service is not None:
            service = entities.Service.from_json(_json=service,
                                                 client_api=self._client_api,
                                                 project=self._project,
                                                 package=self._package)
        return service

    def checkout(self, service: entities.Service = None, service_name=None, service_id=None):
        """
        Check-out a service
        :param service_id:
        :param service_name:
        :param service: Service entity
        :return:
        """
        if service is None:
            service = self.get(service_name=service_name, service_id=service_id)
        self._client_api.state_io.put('service', service.to_json())
        logger.info('Checked out to service {}'.format(service.name))

    ###########
    # methods #
    ###########
    def get(self, service_name=None, service_id=None, checkout=False, fetch=None) -> entities.Service:
        """
        Get service

        :param checkout:
        :param service_name: optional - search by name
        :param service_id: optional - search by id
        :param fetch: optional - fetch entity from platform, default taken from cookie
        :return: Service object
        """
        if fetch is None:
            fetch = self._client_api.fetch_entities

        if service_name is None and service_id is None:
            service = self.__get_from_cache()
            if service is None:
                raise exceptions.PlatformException(
                    error='400',
                    message='Checked out not found, must provide either service id or service name')

        elif fetch:
            if service_id is not None:
                success, response = self._client_api.gen_request(
                    req_type="get",
                    path="/services/{}".format(service_id)
                )
                if not success:
                    raise exceptions.PlatformException(response)
                service = entities.Service.from_json(client_api=self._client_api,
                                                     _json=response.json(),
                                                     package=self._package,
                                                     project=self._project)
            elif service_name is not None:
                filters = entities.Filters(resource=entities.FiltersResource.SERVICE,
                                           field='name',
                                           values=service_name,
                                           use_defaults=False)
                if self._project_id is not None:
                    filters.add(field='projectId', values=self._project_id)
                if self._package is not None:
                    filters.add(field='packageId', values=self._package.id)
                services = self.list(filters=filters)
                if services.items_count > 1:
                    raise exceptions.PlatformException('404', 'More than one service with same name. '
                                                              'Please get services from package/project entity')
                elif services.items_count == 0:
                    raise exceptions.PlatformException('404', 'Service not found: {}.'.format(service_name))
                service = services[0]
            else:
                raise exceptions.PlatformException(
                    error='400',
                    message='Checked out not found, must provide either service id or service name')
        else:
            service = entities.Service.from_json(_json={'id': service_id,
                                                        'name': service_name},
                                                 client_api=self._client_api,
                                                 project=self._project,
                                                 package=self,
                                                 is_fetched=False)

        assert isinstance(service, entities.Service)
        if checkout:
            self.checkout(service=service)
        return service

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Service]:
        jobs = [None for _ in range(len(response_items))]
        pool = self._client_api.thread_pools(pool_name='entity.create')

        # return triggers list
        for i_service, service in enumerate(response_items):
            jobs[i_service] = pool.apply_async(entities.Service._protected_from_json,
                                               kwds={'client_api': self._client_api,
                                                     '_json': service,
                                                     'package': self._package,
                                                     'project': self._project})
        # wait for all jobs
        _ = [j.wait() for j in jobs]
        # get all results
        results = [j.get() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        return miscellaneous.List([r[1] for r in results if r[0] is True])

    def _list(self, filters: entities.Filters):
        url = '/query/FaaS'
        success, response = self._client_api.gen_request(req_type='POST',
                                                         path=url,
                                                         json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)

        return response.json()

    def list(self, filters: entities.Filters = None) -> entities.PagedEntities:
        """
        List project services
        :return:
        """
        # default filters
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.SERVICE)
        if self._project is not None:
            filters.add(field='projectId', values=self._project.id)
        elif self._project_id is not None:
            filters.add(field='projectId', values=self._project_id)
        if self._package is not None:
            filters.add(field='packageId', values=self._package.id)

        # assert type filters
        if not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException('400', 'Unknown filters type')

        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       client_api=self._client_api)
        paged.get_page()
        return paged

    def status(self, service_name=None, service_id=None):
        if service_id is None:
            if service_name is None:
                raise exceptions.PlatformException(error='400',
                                                   message='must input "service_name" or "service_id" to get status')
            service = self.get(service_name=service_name)
            service_id = service.id
        # request
        success, response = self._client_api.gen_request(req_type="get",
                                                         path="/services/{}/status".format(service_id))
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    def pause(self, service_name=None, service_id=None):
        if service_id is None:
            if service_name is None:
                raise exceptions.PlatformException(error='400',
                                                   message='must input "service_name" or "service_id" to pause service')
            service = self.get(service_name=service_name)
            service_id = service.id
        # request
        success, response = self._client_api.gen_request(req_type="post",
                                                         path="/services/{}/stop".format(service_id))
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    def resume(self, service_name=None, service_id=None):
        if service_id is None:
            if service_name is None:
                raise exceptions.PlatformException(error='400',
                                                   message='must input "service_name" or "service_id" to resume')
            service = self.get(service_name=service_name)
            service_id = service.id
        # request
        success, response = self._client_api.gen_request(req_type="post",
                                                         path="/services/{}/resume".format(service_id))
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    def _get_bot_email(self, bot=None):

        if bot is None:
            project = self._project
            if project is None and self._project_id is not None:
                project = repositories.Projects(client_api=self._client_api).get(project_id=self._project_id)

            if project is None:
                raise exceptions.PlatformException(error='2001',
                                                   message='Need project entity or bot to perform this action')
            bots = project.bots.list()
            if len(bots) == 0:
                logger.info('Bot not found for project. Creating a default bot')
                bot = project.bots.create(name='default')
            else:
                bot = bots[0]
                if len(bots) > 1:
                    logger.debug('More than one bot users. Choosing first. email: {}'.format(bots[0].email))

        if isinstance(bot, str):
            bot_email = bot
        elif isinstance(bot, entities.Bot) or isinstance(bot, entities.User):
            bot_email = bot.email
        else:
            raise ValueError('input "bot" must be a str or a Bot type, got: {}'.format(type(bot)))

        return bot_email

    @staticmethod
    def _parse_init_input(init_input):
        if not isinstance(init_input, dict):
            if init_input is None:
                init_input = dict()
            else:
                init_params = dict()
                if not isinstance(init_input, list):
                    init_input = [init_input]
                for param in init_input:
                    if isinstance(param, entities.FunctionIO):
                        init_params.update(param.to_json(resource='service'))
                    else:
                        raise exceptions.PlatformException('400', 'Unknown type of init params')
                init_input = init_params

        return init_input

    def _create(self, service_name=None, package=None, module_name=None, bot=None, revision=None, init_input=None,
                runtime=None, pod_type=None, project_id=None, sdk_version=None, agent_versions=None, verify=True,
                driver_id=None, run_execution_as_process=None, execution_timeout=None, drain_time=None, on_reset=None,
                **kwargs) -> entities.Service:
        """
        Create service entity
        :param verify:
        :param agent_versions:
        :param sdk_version:
        :param project_id:
        :param pod_type:
        :param runtime:
        :param init_input:
        :param driver_id:
        :param bot: bot user that weill run the package
        :param revision: optional - int - package revision - default=latest
        :param package:
        :param service_name:
        :return:
        """
        is_global = kwargs.get('is_global', None)
        jwt_forward = kwargs.get('jwt_forward', None)
        if is_global is not None or jwt_forward is not None:
            logger.warning(
                'Params jwt_forward and is_global are restricted to superuser. '
                'If you are not a superuser this action will not work')

        if agent_versions is None:
            if sdk_version is None:
                sdk_version = __version__
            agent_versions = {
                "dtlpy": sdk_version,
                "verify": verify
            }

        if project_id is None:
            if self._project is None and self._project_id is None:
                raise exceptions.PlatformException('400', 'Please provide project id')
            elif self._project_id is not None:
                project_id = self._project_id
            elif self._project is not None:
                project_id = self._project.id

        if package is None:
            if self._package is None:
                raise exceptions.PlatformException('400', 'Please provide param package')
            package = self._package

        if module_name is None:
            module_name = entities.package_defaults.DEFAULT_PACKAGE_MODULE_NAME
        if service_name is None:
            service_name = 'default-service'

        # payload
        payload = {'name': service_name,
                   'projectId': project_id,
                   'packageId': package.id,
                   'initParams': self._parse_init_input(init_input=init_input),
                   'botUserName': self._get_bot_email(bot=bot),
                   'versions': agent_versions,
                   'moduleName': module_name,
                   'driverId': driver_id}

        # revision
        if isinstance(revision, int):
            payload['packageRevision'] = revision

        if runtime is not None:
            if isinstance(runtime, entities.KubernetesRuntime):
                runtime = runtime.to_json()
            payload['runtime'] = runtime
        else:
            payload['runtime'] = {'gpu': False,
                                  'numReplicas': 1}

        if pod_type is None:
            if 'gpu' in payload['runtime'] and payload['runtime']['gpu']:
                pod_type = 'gpu-k80-s'
            else:
                pod_type = 'regular-s'
        elif 'podType' not in payload['runtime'] and payload['runtime']['podType'] != pod_type:
            raise exceptions.PlatformException('400', 'Different pod_types given in param and in runtime.\n'
                                                      'pod_type param = {}\n'
                                                      'pod_type in runtime = {}'.format(pod_type,
                                                                                        payload['runtime']['podType']))

        if is_global is not None:
            payload['global'] = is_global

        if jwt_forward is not None:
            payload['useUserJwt'] = jwt_forward

        if 'podType' not in payload['runtime']:
            payload['runtime']['podType'] = pod_type

        if run_execution_as_process is not None:
            payload['runExecutionAsProcess'] = run_execution_as_process

        if drain_time is not None:
            payload['drainTime'] = drain_time

        if on_reset is not None:
            payload['onReset'] = on_reset

        if execution_timeout is not None:
            payload['executionTimeout'] = execution_timeout

        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/services',
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.Service.from_json(_json=response.json(),
                                          client_api=self._client_api,
                                          package=package,
                                          project=self._project)

    def delete(self, service_name=None, service_id=None):
        """
        Delete Service object

        :param service_id: by id
        :param service_name: by name
        :return: True
        """
        # get bby name
        if service_id is None:
            if service_name is None:
                raise exceptions.PlatformException('400', 'Must provide either service id or service name')
            else:
                service_id = self.get(service_name=service_name).id

        # request
        success, response = self._client_api.gen_request(
            req_type="delete",
            path="/services/{}".format(service_id)
        )
        if not success:
            raise exceptions.PlatformException(response)
        return True

    def update(self, service: entities.Service) -> entities.Service:
        """
        Update Service changes to platform
        :param service: Service entity
        :return: Service entity
        """
        assert isinstance(service, entities.Service)

        # payload
        payload = service.to_json()

        # request
        success, response = self._client_api.gen_request(req_type='patch',
                                                         path='/services/{}'.format(service.id),
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        if self._package is not None:
            package = self._package
        else:
            package = service._package

        return entities.Service.from_json(_json=response.json(),
                                          client_api=self._client_api,
                                          package=package,
                                          project=self._project)

    def log(self, service, size=None, checkpoint=None, start=None, end=None, follow=False, text=None,
            execution_id=None, function_name=None, replica_id=None, system=False, view=True):
        """
        Get service logs

        :param text:
        :param view:
        :param system:
        :param end: iso format time
        :param start: iso format time
        :param checkpoint:
        :param follow: filters
        :param execution_id: follow
        :param function_name: execution_id
        :param replica_id: function_name
        :param size:
        :param service: Service entity
        :return: Service entity
        """
        assert isinstance(service, entities.Service)

        payload = {
            'direction': 'asc',
            'follow': follow,
            'system': system
        }

        if size is not None:
            payload['size'] = size

        if execution_id is not None:
            payload['executionId'] = execution_id

        if function_name is not None:
            payload['functionName'] = function_name

        if text is not None:
            payload['text'] = text

        if replica_id is not None:
            payload['replicaId'] = replica_id

        if checkpoint is not None:
            payload['checkpoint'] = checkpoint

        if start is not None:
            payload['start'] = start
        else:
            payload['start'] = datetime.datetime(datetime.date.today().year,
                                                 datetime.date.today().month,
                                                 datetime.date.today().day,
                                                 0,
                                                 0,
                                                 0).isoformat()

        if end is not None:
            payload['end'] = end

        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/services/{}/logs'.format(service.id),
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        log = ServiceLog(_json=response.json(),
                         service=service,
                         services=self,
                         start=payload['start'],
                         follow=follow,
                         execution_id=execution_id,
                         function_name=function_name,
                         replica_id=replica_id,
                         system=system)

        if view:
            log.view()
        else:
            return log

    def execute(self,
                service: entities.Service = None,
                service_id: str = None,
                service_name: str = None,
                sync: bool = False,
                function_name: str = None,
                stream_logs: bool = False,
                execution_input=None,
                resource=None,
                item_id=None,
                dataset_id=None,
                annotation_id=None,
                project_id=None,
                ) -> entities.Execution:
        if service is None:
            service = self.get(service_id=service_id, service_name=service_name)
        execution = repositories.Executions(service=service,
                                            client_api=self._client_api,
                                            project=self._project).create(service_id=service.id,
                                                                          sync=sync,
                                                                          execution_input=execution_input,
                                                                          function_name=function_name,
                                                                          resource=resource,
                                                                          item_id=item_id,
                                                                          dataset_id=dataset_id,
                                                                          annotation_id=annotation_id,
                                                                          project_id=project_id,
                                                                          stream_logs=stream_logs)
        return execution

    def deploy(self,
               service_name: str = None,
               package: entities.Package = None,
               bot: Union[entities.Bot, str] = None,
               revision: int = None,
               init_input: Union[List[entities.FunctionIO], entities.FunctionIO, dict] = None,
               runtime: Union[entities.KubernetesRuntime, dict] = None,
               pod_type: str = None,
               sdk_version: str = None,
               agent_versions: dict = None,
               verify: bool = True,
               checkout: bool = False,
               module_name: str = None,
               project_id: str = None,
               driver_id: str = None,
               func: Callable = None,
               run_execution_as_process: bool = None,
               execution_timeout: int = None,
               drain_time: int = None,
               on_reset: str = None,
               is_staticmethod: bool = False,
               imports: List[dict] = None,
               **kwargs) -> entities.Service:
        """
        Deploy service

        :param on_reset:
        :param drain_time:
        :param execution_timeout:
        :param run_execution_as_process:
        :param func:
        :param project_id:
        :param module_name:
        :param checkout:
        :param verify:
        :param pod_type:
        :param service_name: name
        :param package: package entity
        :param bot: bot email
        :param revision: version
        :param init_input: config to run at startup
        :param runtime: runtime resources
        :param driver_id:
        :param agent_versions: - dictionary - - optional -versions of sdk, agent runner and agent proxy
        :param sdk_version:  - optional - string - sdk version
        :param imports: optional - if param func is given - dictionary of imports, example: [{'module': 'os'}, {'from': 'dictdiffer', 'module': 'diff'}, 'module': 'numpy', 'as': 'np']
        :param is_staticmethod: optional - if param func is given - True/False
        :return:
        """
        if service_name is None:
            package = package if package is not None else self._package
            if package is not None:
                service_name = package.name
            else:
                service_name = 'default-service'
            logger.warning('service_name not provided, using: {} by default'.format(service_name))

        if func is not None:
            return self.__deploy_function(name=service_name, project=self._project,
                                          func=func, is_staticmethod=is_staticmethod, imports=imports)

        if init_input is not None and not isinstance(init_input, dict):
            if not isinstance(init_input, list):
                init_input = [init_input]

            if len(init_input) > 0 and isinstance(init_input[0], entities.FunctionIO):
                params = dict()
                for i_param, param in enumerate(init_input):
                    params[param.name] = param.value
                init_input = params
            elif len(init_input) == 0:
                init_input = None
            else:
                raise exceptions.PlatformException(
                    error='400',
                    message='Unknown init_input type. expecting list or dict, got: {}'.format(type(init_input)))

        if project_id is None:
            if self._project is not None:
                project_id = self._project.id
            else:
                project_id = self._project_id

        filters = entities.Filters(resource=entities.FiltersResource.SERVICE)
        filters.add(field='name', values=service_name)
        if project_id is not None:
            filters.add(field='projectId', values=project_id)
        services = self.list(filters=filters)
        if services.items_count > 1:
            raise exceptions.PlatformException('400',
                                               'More than 1 service by this name are associated with this user. '
                                               'Please provide project_id')
        elif services.items_count > 0:
            service = services[0]
            if runtime is not None:
                service.runtime = runtime
            if init_input is not None:
                service.init_input = init_input
            if revision is not None:
                service.package_revision = revision
            if agent_versions is not None:
                service.versions = agent_versions
            if driver_id is not None:
                service.driver_id = driver_id
            service = self.update(service=service)
        else:
            service = self._create(service_name=service_name,
                                   package=package,
                                   bot=bot,
                                   revision=revision,
                                   init_input=init_input,
                                   runtime=runtime,
                                   pod_type=pod_type,
                                   sdk_version=sdk_version,
                                   agent_versions=agent_versions,
                                   verify=verify,
                                   module_name=module_name,
                                   driver_id=driver_id,
                                   jwt_forward=kwargs.get('jwt_forward', None),
                                   is_global=kwargs.get('is_global', None),
                                   run_execution_as_process=run_execution_as_process,
                                   execution_timeout=execution_timeout,
                                   drain_time=drain_time,
                                   on_reset=on_reset)
        if checkout:
            self.checkout(service=service)
        return service

    @staticmethod
    def __get_import_string(imports: List[dict]):
        imports_string = ''
        if imports is not None:
            for imprt in imports:
                as_string = ''
                from_string = ''
                import_from = imprt.get('from', None)
                import_as = imprt.get('as', None)
                import_module = imprt.get('module', None)
                if import_from is not None:
                    from_string = 'from {} '.format(import_from)
                if import_as is not None:
                    as_string = ' as {}'.format(import_as)
                imports_string += '{}import {}{}\n'.format(from_string, import_module, as_string)
        return imports_string

    @staticmethod
    def __get_inputs(func):
        method = inspect.signature(func)
        params = list(method.parameters)
        inpts = list()
        for arg in params:
            if arg == 'item':
                inpt_type = 'Item'
            elif arg == 'dataset':
                inpt_type = 'Dataset'
            else:
                inpt_type = 'Json'
            inpts.append(entities.FunctionIO(type=inpt_type, name=arg))
        return inpts

    def __deploy_function(self,
                          name: str,
                          func: Callable,
                          project: entities.Project,
                          imports: List[dict] = None,
                          is_staticmethod: bool = True) -> entities.Service:
        package_dir = tempfile.mkdtemp()
        imports_string = self.__get_import_string(imports=imports)
        main_file = os.path.join(package_dir, entities.package_defaults.DEFAULT_PACKAGE_ENTRY_POINT)
        with open(assets.paths.PARTIAL_MAIN_FILEPATH, 'r') as f:
            main_string = f.read()
        lines = inspect.getsourcelines(func)
        method_func_string = "".join(lines[0])

        with open(main_file, 'w') as f:
            if is_staticmethod:
                f.write('{}\n{}\n    @staticmethod\n    {}'.format(imports_string, main_string,
                                                                   method_func_string.replace('\n', '\n    ')))
            else:
                f.write('{}\n{}\n    {}'.format(imports_string, main_string,
                                                method_func_string.replace('\n', '\n    ')))

        function = entities.PackageFunction(name=func.__name__, inputs=self.__get_inputs(func=func))
        module = entities.PackageModule(functions=function,
                                        entry_point=entities.package_defaults.DEFAULT_PACKAGE_ENTRY_POINT)
        packages = repositories.Packages(client_api=self._client_api, project=project)
        return packages.push(src_path=package_dir,
                             package_name=name,
                             checkout=True,
                             modules=module).deploy(service_name=name)

    def deploy_from_local_folder(self, cwd=None, service_file=None, bot=None, checkout=False) -> entities.Service:
        """
        Deploy from local folder
        :param checkout:
        :param bot:
        :return:

        :param cwd: optional - package working directory. Default=cwd
        :param service_file: optional - service file. Default=None
        """
        # get cwd and service.json path
        if cwd is None:
            cwd = os.getcwd()
        if service_file is None:
            service_file = os.path.join(cwd, assets.paths.SERVICE_FILENAME)

        # load service json
        if os.path.isfile(service_file):
            with open(service_file, 'r') as f:
                service_json = json.load(f)
            service_triggers = service_json.get('triggers', list())
        else:
            raise exceptions.PlatformException(error='400',
                                               message='Could not find service.json in path: {}'.format(cwd))

        # get package
        package_name = service_json.get('packageName', None)
        packages = repositories.Packages(client_api=self._client_api, project=self._project)
        if package_name is None:
            package = packages.get()
        else:
            package = packages.get(package_name=package_name)

        name = service_json.get('name', None)
        revision = service_json.get('revision', package.version)
        init_input = service_json.get('initParams', dict())
        runtime = service_json.get('runtime', dict())
        sdk_version = service_json.get('version', None)
        agent_versions = service_json.get('versions', None)
        verify = service_json.get('verify', True)
        module_name = service_json.get('module_name', None)
        run_execution_as_process = service_json.get('run_execution_as_process', None)
        execution_timeout = service_json.get('execution_timeout', None)
        drain_time = service_json.get('drain_time', None)
        on_reset = service_json.get('on_reset', None)

        service = self.deploy(bot=bot,
                              service_name=name,
                              package=package,
                              revision=revision,
                              runtime=runtime,
                              init_input=init_input,
                              sdk_version=sdk_version,
                              agent_versions=agent_versions,
                              verify=verify,
                              checkout=checkout,
                              run_execution_as_process=run_execution_as_process,
                              execution_timeout=execution_timeout,
                              drain_time=drain_time,
                              on_reset=on_reset,
                              module_name=module_name)

        triggers = repositories.Triggers(client_api=self._client_api, project=self._project)
        for trigger in service_triggers:
            name = trigger.get('name', None)
            filters = trigger.get('filter', dict())
            resource = trigger['resource']
            actions = trigger.get('actions', list())
            active = trigger.get('active', True)
            execution_mode = trigger.get('executionMode', None)
            function_name = trigger.get('function', None)

            triggers.create(service_id=service.id,
                            name=name,
                            filters=filters,
                            resource=resource,
                            actions=actions,
                            active=active,
                            execution_mode=execution_mode,
                            function_name=function_name)

        logging.debug('Successfully deployed!')
        return service

    def deploy_pipeline(self, service_json_path=None, project=None, bot=None):
        """
        Deploy pipeline

        :param bot:
        :param project:
        :param service_json_path:
        :return: True
        """
        # project
        if project is None:
            project = self._project

        # get service file
        if service_json_path is None:
            service_json_path = os.getcwd()

        if not service_json_path.endswith('.json'):
            service_json_path = os.path.join(service_json_path, 'service.json')
            if not os.path.isfile(service_json_path):
                raise exceptions.PlatformException('404', 'File not exist: {}'.format(service_json_path))

        # get existing project's services and triggers
        project_triggers = {trigger.name: trigger for trigger in project.triggers.list()}
        project_services = {service.name: service for service in project.services.list()}
        project_packages = {package.name: package for package in project.packages.list()}

        # load file
        with open(service_json_path, 'r') as f:
            service_json = json.load(f)

        # build
        for service_input in service_json:
            # get package
            if service_input['package'] in project_packages:
                package = project_packages[service_input['package']]
            else:
                raise exceptions.PlatformException('404', 'Package not found, package name: {}'.format(
                    service_input['package']))

            sdk_version = service_json.get('version', None)
            agent_versions = service_json.get('versions', None)
            verify = service_json.get('verify', True)
            runtime = service_input.get('runtime', None)
            init_input = service_input.get('initParams', None)
            module_name = service_input.get('module_name', None)
            driver_id = service_input.get('driverId', None)
            run_execution_as_process = service_json.get('run_execution_as_process', None)
            execution_timeout = service_json.get('execution_timeout', None)
            drain_time = service_json.get('drain_time', None)
            on_reset = service_json.get('on_reset', None)
            # create or update service
            if service_input['name'] in project_services:
                service = project_services[service_input['name']]
                service.runtime = runtime
                service.init_input = init_input
                service = project.services.update(service=service)
                service.version = sdk_version
                service.versions = agent_versions
                service.verify = verify
                service.driver_id = driver_id
                project_services[service.name] = service

            else:
                service = self._create(package=package,
                                       bot=bot,
                                       service_name=service_input['name'],
                                       runtime=runtime,
                                       init_input=init_input,
                                       agent_versions=agent_versions,
                                       sdk_version=sdk_version,
                                       verify=verify,
                                       module_name=module_name,
                                       run_execution_as_process=run_execution_as_process,
                                       execution_timeout=execution_timeout,
                                       drain_time=drain_time,
                                       on_reset=on_reset,
                                       driver_id=driver_id)
                project_services[service.name] = service

            service_triggers = {trigger: project_triggers[trigger] for trigger in project_triggers if
                                project_triggers[trigger].service_id == service.id}

            # create or update triggers
            if 'triggers' in service_input:
                for trigger_input in service_input['triggers']:
                    if trigger_input['name'] in service_triggers:
                        trigger = project_triggers[trigger_input['name']]
                        trigger.resource = trigger_input.get('resource', trigger.resource)
                        trigger.active = trigger_input.get('active', trigger.active)
                        trigger.actions = trigger_input.get('actions', trigger.actions)
                        trigger.filters = trigger_input.get('filters', trigger.filters)
                        trigger.execution_mode = trigger_input.get('executionMode', trigger.execution_mode)
                        trigger.function_name = trigger_input.get('function', None)
                        trigger = trigger.update()
                        project_triggers[trigger.name] = trigger
                    else:
                        trigger = project.triggers.create(service_ids=[service.id],
                                                          resource=trigger_input.get('resource', None),
                                                          active=trigger_input.get('active', None),
                                                          actions=trigger_input.get('actions', None),
                                                          filters=trigger_input.get('filters', None),
                                                          function_name=trigger_input.get('function', None),
                                                          execution_mode=trigger_input.get('executionMode', None))
                        project_triggers[trigger.name] = trigger

        print('File deployed successfully!')
        return True

    def tear_down(self, service_json_path=None, project=None):
        """
        Tear down a pipeline

        :param project:
        :param service_json_path:
        :return:
        """
        # project
        if project is None:
            project = self._project

        # get service file
        if service_json_path is None:
            service_json_path = os.getcwd()

        if not service_json_path.endswith('.json'):
            service_json_path = os.path.join(service_json_path, 'service.json')
            if not os.path.isfile(service_json_path):
                raise exceptions.PlatformException('404', 'File not exist: {}'.format(service_json_path))

        # get existing project's services and triggers
        project_triggers = {trigger.name: trigger for trigger in project.triggers.list()}
        project_services = {service.name: service for service in project.services.list()}

        with open(service_json_path, 'r') as f:
            service_json = json.load(f)

        # tear down
        for service_input in service_json:
            # delete service
            if service_input['name'] in project_services:
                service = project_services[service_input['name']]
                service.delete()

            # delete triggers
            if 'triggers' in service_input:
                for trigger_input in service_input['triggers']:
                    if trigger_input['name'] in project_triggers:
                        trigger = project_triggers[trigger_input['name']]
                        trigger.delete()

        print('File torn down successfully!')
        return True

    @staticmethod
    def generate_services_json(path=None):
        if path is None:
            path = os.getcwd()
        path = os.path.join(path, assets.paths.SERVICE_FILENAME)
        with open(assets.paths.ASSETS_SERVICE_FILEPATH, 'r') as f:
            service_file = json.load(f)
        with open(path, 'w') as f:
            json.dump(service_file, f, indent=2)
        return path


class ServiceLog:
    """
    Service Log
    """

    def __init__(self,
                 _json: dict,
                 service: entities.Service,
                 services: Services,
                 start=None,
                 follow=None,
                 execution_id=None,
                 function_name=None,
                 replica_id=None,
                 system=False):

        self.logs = _json.get('logs', dict())
        self.checkpoint = _json.get('checkpoint', None)
        self.stop = _json.get('stop', False)
        self.service = service
        self.services = services
        self.start = start
        self.follow = follow
        self.execution_id = execution_id
        self.function_name = function_name
        self.replica_id = replica_id
        self.system = system

    def get_next_log(self):
        log = self.services.log(service=self.service,
                                checkpoint=self.checkpoint,
                                start=self.start,
                                follow=self.follow,
                                execution_id=self.execution_id,
                                function_name=self.function_name,
                                replica_id=self.replica_id,
                                system=self.system,
                                view=False)

        self.logs = log.logs
        self.checkpoint = log.checkpoint
        self.stop = log.stop

    def view(self):
        try:
            for log in self:
                print(log)
        except KeyboardInterrupt:
            return

    def __iter__(self):
        while not self.stop:
            for log in self.logs:
                yield '{}: {}'.format(log.get('timestamp', self.start), log.get('message', '').strip())
            self.get_next_log()



import mimetypes
import traceback
import datetime
import webvtt
import logging
import attr
import json
import os

import numpy as np
from PIL import Image

from .. import entities, PlatformException

logger = logging.getLogger(name=__name__)


@attr.s
class AnnotationCollection(entities.BaseEntity):
    """
        Collection of Annotation entity
    """
    item = attr.ib()
    annotations = attr.ib()

    @annotations.default
    def set_annotations(self):
        return list()

    def __iter__(self):
        for annotation in self.annotations:
            yield annotation

    def __getitem__(self, index: int):
        return self.annotations[index]

    def __len__(self):
        return len(self.annotations)

    def add(self,
            annotation_definition,
            object_id=None,
            frame_num=None,
            end_frame_num=None,
            start_time=None,
            end_time=None,
            automated=True,
            fixed=True,
            metadata=None,
            parent_id=None,
            model_info=None):
        """
        Add annotations to collection

        :param annotation_definition: dl.Polygon, dl.Segmentation, dl.Point, dl.Box etc
        :param object_id: Object id (any id given by user). If video - must input to match annotations between frames
        :param frame_num: video only, number of frame
        :param end_frame_num: video only, the end frame of the annotation
        :param start_time: video only, start time of the annotation
        :param end_time: video only, end time of the annotation
        :param fixed: video only, mark frame as fixed
        :param parent_id: set a parent for this annotation (parent annotation ID)
        :param automated:
        :param metadata: optional- metadata dictionary for annotation
        :param model_info: optional - set model on annotation {'name',:'', 'confidence':0}
        :return:
        """
        if model_info is not None:
            if not isinstance(model_info, dict) or 'name' not in model_info or 'confidence' not in model_info:
                raise ValueError('"model_info" must be a dict with keys: "name" and "confidence"')
            if metadata is None:
                metadata = dict()
            if 'user' not in metadata:
                metadata['user'] = dict()
            metadata['user']['model'] = {'name': model_info['name'],
                                         'confidence': float(model_info['confidence'])}

        # to support list of definitions with same parameters
        if not isinstance(annotation_definition, list):
            annotation_definition = [annotation_definition]

        for single_definition in annotation_definition:
            if object_id is None:
                # add new annotation to list
                annotation = entities.Annotation.new(item=self.item,
                                                     annotation_definition=single_definition,
                                                     frame_num=frame_num,
                                                     automated=automated,
                                                     metadata=metadata,
                                                     parent_id=parent_id)
                self.annotations.append(annotation)
                matched_ind = len(self.annotations) - 1
            else:
                # find matching element_id
                matched_ind = [i_annotation
                               for i_annotation, annotation in enumerate(self.annotations)
                               if annotation.object_id == object_id]
                if len(matched_ind) == 0:
                    # no matching object id found - create new one
                    annotation = entities.Annotation.new(item=self.item,
                                                         annotation_definition=single_definition,
                                                         frame_num=frame_num,
                                                         automated=automated,
                                                         metadata=metadata,
                                                         object_id=object_id,
                                                         parent_id=parent_id)
                    self.annotations.append(annotation)
                    matched_ind = len(self.annotations) - 1
                elif len(matched_ind) == 1:
                    matched_ind = matched_ind[0]
                else:
                    raise PlatformException(error='400',
                                            message='more than one annotation with same object id: {}'.format(
                                                object_id))

            #  add frame if exists
            if frame_num is not None or start_time is not None:
                self.annotations[matched_ind].add_frames(annotation_definition=single_definition,
                                                         frame_num=frame_num,
                                                         end_frame_num=end_frame_num,
                                                         start_time=start_time,
                                                         end_time=end_time,
                                                         fixed=fixed)

    ############
    # Plotting #
    ############
    def show(self, image=None, thickness=None, with_text=False, height=None, width=None,
             annotation_format=entities.ViewAnnotationOptions.MASK,
             label_instance_dict=None):
        """
            Show annotations according to annotation_format

        :param image: empty or image to draw on
        :param height: height
        :param width: width
        :param thickness: line thickness
        :param with_text: add label to annotation
        :param annotation_format: how to show thw annotations. options: dl.ViewAnnotationOptions.list()
        :param label_instance_dict: instance label map {'Label': 1, 'More': 2}
        :return: ndarray of the annotations
        """
        # if 'video' in self.item.mimetype and (annotation_format != 'json' or annotation_format != ['json']):
        #     raise PlatformException('400', 'Cannot show mask or instance of video item')
        # height/weight
        try:
            import cv2
        except ImportError:
            logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise

        if height is None:
            if self.item.height is None:
                raise PlatformException('400', 'Height must be provided')
            height = self.item.height
        if width is None:
            if self.item.width is None:
                raise PlatformException('400', 'Width must be provided')
            width = self.item.width

        if annotation_format == entities.ViewAnnotationOptions.MASK:
            # create an empty mask
            if image is None:
                mask = np.zeros((height, width, 4), dtype=np.uint8)
            else:
                if len(image.shape) == 2:
                    # image is gray
                    mask = cv2.cvtColor(image, cv2.COLOR_GRAY2RGBA)
                elif image.shape[2] == 3:
                    mask = cv2.cvtColor(image, cv2.COLOR_RGB2RGBA)
                else:
                    raise PlatformException(error='1001',
                                            message='Unknown image shape. expected depth: gray or RGB. got: {}'.format(
                                                image.shape))
        elif annotation_format == entities.ViewAnnotationOptions.INSTANCE:
            if image is None:
                # create an empty mask
                mask = np.zeros((height, width), dtype=np.uint8)
            else:
                if len(image.shape) != 2:
                    raise PlatformException(error='1001',
                                            message='Image shape must be 2d array when trying to draw instance on image')
                mask = image
            # create a dictionary of labels and ids
            if label_instance_dict is None:
                label_instance_dict = self.item.dataset.instance_map
        elif annotation_format == entities.ViewAnnotationOptions.OBJECT_ID:
            if image is None:
                # create an empty mask
                mask = np.zeros((height, width), dtype=np.uint8)
            else:
                if len(image.shape) != 2:
                    raise PlatformException(error='1001',
                                            message='Image shape must be 2d array when trying to draw instance on image')
                mask = image
        else:
            raise PlatformException(error='1001',
                                    message='unknown annotations format: "{}". known formats: "{}"'.format(
                                        annotation_format, '", "'.join(entities.ViewAnnotationOptions.list())))

        #############
        # gor over all annotations and put the id where the annotations is
        for annotation in self.annotations:
            if annotation_format == entities.ViewAnnotationOptions.MASK:
                color = None
            elif annotation_format == entities.ViewAnnotationOptions.INSTANCE:
                # if label not in dataset label - put it as background
                color = label_instance_dict.get(annotation.label, 0)
            elif annotation_format == entities.ViewAnnotationOptions.OBJECT_ID:
                if annotation.object_id is None:
                    raise PlatformException(error='1001',
                                            message='Try to show object_id but annotation has no value. annotation id: {}'.format(
                                                annotation.id))
                color = annotation.object_id
            else:
                raise PlatformException('404',
                                        'unknown annotations format: {}. known formats: "{}"'.format(
                                            annotation_format, '", "'.join(entities.ViewAnnotationOptions.list())))
            # get the mask of the annotation
            mask = annotation.show(thickness=thickness,
                                   color=color,
                                   with_text=with_text,
                                   height=height,
                                   width=width,
                                   annotation_format=annotation_format,
                                   image=mask)

        return mask

    def download(self, filepath, img_filepath=None, annotation_format=entities.ViewAnnotationOptions.MASK, height=None,
                 width=None, thickness=1,
                 with_text=False):
        """
            Save annotations to file

        :param filepath: path to save annotation
        :param img_filepath: img file path - needed for img_mask
        :param annotation_format:
        :param height:
        :param width:
        :param thickness:
        :param with_text:
        :return:
        """
        dir_name, ex = os.path.splitext(filepath)

        if annotation_format == entities.ViewAnnotationOptions.JSON:
            if not ex:
                filepath = '{}/{}.json'.format(dir_name, os.path.splitext(self.item.name)[0])
            _json = {'_id': self.item.id,
                     'filename': self.item.filename}
            annotations = list()
            for ann in self.annotations:
                annotations.append(ann.to_json())
            _json['annotations'] = annotations
            with open(filepath, 'w+') as f:
                json.dump(_json, f, indent=2)
        elif annotation_format in [entities.ViewAnnotationOptions.MASK,
                                   entities.ViewAnnotationOptions.INSTANCE,
                                   entities.ViewAnnotationOptions.ANNOTATION_ON_IMAGE]:
            if not ex:
                filepath = '{}/{}.png'.format(dir_name, os.path.splitext(self.item.name)[0])
            image = None
            if annotation_format == entities.ViewAnnotationOptions.ANNOTATION_ON_IMAGE:
                annotation_format = entities.ViewAnnotationOptions.MASK
                image = np.asarray(Image.open(img_filepath))
            mask = self.show(image=image,
                             thickness=thickness,
                             with_text=with_text,
                             height=height,
                             width=width,
                             annotation_format=annotation_format)
            img = Image.fromarray(mask.astype(np.uint8))
            img.save(filepath)
        elif annotation_format == entities.ViewAnnotationOptions.VTT:
            if not ex:
                filepath = '{}/{}.vtt'.format(dir_name, os.path.splitext(self.item.name)[0])
            annotations_dict = [{'start_time': annotation.start_time,
                                 'end_time': annotation.end_time,
                                 'text': annotation.coordinates['text']} for annotation in self.annotations if
                                annotation.type in ['subtitle']]
            sorted_by_start_time = sorted(annotations_dict, key=lambda i: i['start_time'])
            vtt = webvtt.WebVTT()
            for ann in sorted_by_start_time:
                s = str(datetime.timedelta(seconds=ann['start_time']))
                if len(s.split('.')) == 1:
                    s += '.000'
                e = str(datetime.timedelta(seconds=ann['end_time']))
                if len(e.split('.')) == 1:
                    e += '.000'
                caption = webvtt.Caption(
                    '{}'.format(s),
                    '{}'.format(e),
                    '{}'.format(ann['text'])
                )
                vtt.captions.append(caption)
            vtt.save(filepath)
        else:
            raise PlatformException(error="400", message="Unknown annotation option: {}".format(annotation_format))
        return filepath

    ############
    # Platform #
    ############
    def update(self, system_metadata=True):
        if self.item is None:
            raise PlatformException('400', 'missing item to perform platform update')
        return self.item.annotations.update(annotations=self.annotations, system_metadata=system_metadata)

    def delete(self):
        if self.item is None:
            raise PlatformException('400', 'missing item to perform platform delete')
        return [self.item.annotations.delete(annotation_id=annotation.id) for annotation in self.annotations]

    def upload(self):
        if self.item is None:
            raise PlatformException('400', 'missing item to perform platform upload')
        return self.item.annotations.upload(self.annotations)

    @staticmethod
    def _json_to_annotation(item: entities.Item, w_json: dict, is_video=None, fps=25, item_metadata=None):
        try:
            annotation = entities.Annotation.from_json(_json=w_json,
                                                       fps=fps,
                                                       item_metadata=item_metadata,
                                                       is_video=is_video,
                                                       item=item)
            status = True
        except Exception:
            annotation = traceback.format_exc()
            status = False
        return status, annotation

    @classmethod
    def from_json(cls, _json: list, item=None, is_video=None, fps=25, height=None, width=None):
        if item is None:
            if isinstance(_json, dict):
                metadata = _json.get('metadata', dict())
                system_metadata = metadata.get('system', dict())
                if is_video is None:
                    if 'mimetype' in system_metadata:
                        is_video = 'video' in system_metadata['mimetype']
                    elif 'filename' in _json:
                        ext = os.path.splitext(_json['filename'])[-1]
                        try:
                            is_video = 'video' in mimetypes.types_map[ext.lower()]
                        except Exception:
                            logger.info("Unknown annotation's item type. Default item type is set to: image")
                    else:
                        logger.info("Unknown annotation's item type. Default item type is set to: image")
                if is_video:
                    fps = system_metadata.get('fps', fps)
                    ffmpeg_info = system_metadata.get('ffmpeg', dict())
                    height = ffmpeg_info.get('height', None)
                    width = ffmpeg_info.get('width', None)
                else:
                    fps = 0
                    height = system_metadata.get('height', None)
                    width = system_metadata.get('width', None)
        else:
            fps = item.fps
            height = item.height
            width = item.width

        item_metadata = {
            'fps': fps,
            'height': height,
            'width': width
        }

        if 'annotations' in _json:
            _json = _json['annotations']

        results = [None for _ in range(len(_json))]
        for i_json, single_json in enumerate(_json):
            results[i_json] = cls._json_to_annotation(item=item,
                                                      fps=fps,
                                                      item_metadata=item_metadata,
                                                      is_video=is_video,
                                                      w_json=single_json)
        # log errors
        _ = [logger.warning(j[1]) for j in results if j[0] is False]

        # return good jobs
        annotations = [j[1] for j in results if j[0] is True]

        # sort
        if is_video:
            annotations.sort(key=lambda x: x.start_frame)
        else:
            annotations.sort(key=lambda x: x.label)

        return cls(annotations=annotations, item=item)

    def from_vtt_file(self, filepath):
        for caption in webvtt.read(filepath):
            h, m, s = caption.start.split(':')
            start_time = datetime.timedelta(hours=float(h), minutes=float(m), seconds=float(s)).total_seconds()
            h, m, s = caption.end.split(':')
            end_time = datetime.timedelta(hours=float(h), minutes=float(m), seconds=float(s)).total_seconds()

            start_frame = round(start_time * self.item.fps)
            annotation_definition = entities.Subtitle(text=caption.text, label='Text')
            annotation = entities.Annotation.new(
                annotation_definition=annotation_definition,
                frame_num=start_frame,
                item=self.item,
                start_time=start_time)

            annotation.add_frames(annotation_definition=annotation_definition,
                                  frame_num=start_frame, end_time=end_time)

            self.annotations.append(annotation)

    def from_instance_mask(self, mask, instance_map=None):
        if instance_map is None:
            instance_map = self.item.dataset.instance_map
        # go over all instance ids
        for label, instance_id in instance_map.items():
            # find a binary mask per instance
            class_mask = instance_id == mask
            if not np.any(class_mask):
                continue
            # add the binary mask to the annotation builder
            self.add(annotation_definition=entities.Segmentation(geo=class_mask, label=label))

    def to_json(self):

        if self.item is None:
            item_id = None
            item_name = None
        else:
            item_id = self.item.id
            item_name = self.item.filename

        _json = {
            "_id": item_id,
            "filename": item_name,
            'annotations': [annotation.to_json() for annotation in self.annotations]
        }

        return _json

    #########################
    # For video annotations #
    #########################
    def get_frame(self, frame_num):
        frame_collection = AnnotationCollection(item=self.item)
        for annotation in self.annotations:
            if frame_num in annotation.frames:
                annotation.set_frame(frame=frame_num)
                frame_collection.annotations.append(annotation)
        return frame_collection

    def video_player(self):
        from ..utilities.videos.video_player import VideoPlayer
        _ = VideoPlayer(dataset_id=self.item.dataset.id,
                        item_id=self.item.id)



import dtlpy as dl
import logging

logger = logging.getLogger(name=__name__)


class ServiceRunner(dl.BaseServiceRunner):
    """
    Package runner class

    """

    def __init__(self, **kwargs):
        """
        Init package attributes here

        :param kwargs: config params
        :return:
        """



#! /usr/bin/env python3
# This file is part of DTLPY.
#
# DTLPY is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DTLPY is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with DTLPY.  If not, see <http://www.gnu.org/licenses/>.
from . import package_defaults
from .annotation_definitions import BaseAnnotationDefinition
from .base_entity import BaseEntity
from .item import Item, Modality, ModalityTypeEnum, ItemStatus
from .links import Link, ItemLink, UrlLink, LinkTypeEnum
from .trigger import Trigger, TriggerResource, TriggerAction, TriggerExecutionMode
from .project import Project
from .artifact import Artifact
from .dataset import Dataset
from .codebase import Codebase
from .annotation import Annotation, FrameAnnotation, ViewAnnotationOptions
from .annotation_collection import AnnotationCollection
from .paged_entities import PagedEntities
from .filters import Filters, FiltersKnownFields, FiltersResource, FiltersOperations, FiltersMethod, \
    FiltersOrderByDirection
from .recipe import Recipe
from .ontology import Ontology
from .annotation_definitions import Box, Point, Segmentation, Polygon, Ellipse, Classification, Subtitle, \
    Polyline, Comparison, UndefinedAnnotationType, Note
from .label import Label
from .package import Package
from .package_module import PackageModule
from .package_function import PackageFunction, FunctionIO, PackageInputType
from .time_series import TimeSeries
from .service import Service, KubernetesAutuscalerType, KubernetesRabbitmqAutoscaler, KubernetesAutoscaler, \
    InstanceCatalog, KubernetesRuntime
from .execution import Execution, ExecutionStatus
from .assignment import Assignment, Workload, WorkloadUnit
from .task import Task
from .directory_tree import DirectoryTree
from .similarity import Similarity, MultiView, SimilarityItem, MultiViewItem, SimilarityTypeEnum
from .user import User
from .bot import Bot
from .webhook import Webhook, HttpMethod
from .model import Model
from .snapshot import Snapshot
